#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
from openfermion.measurements import partition_iterator

MAX_LOOPS = 1e6


def pair_within(labels: list) -> list:
    """
    Generates pairings of labels that contain each pair at least once.

    A pairing of a list is a set of pairs of list elements. E.g. a pairing of

    labels = [1, 2, 3, 4, 5, 6, 7, 8]

    could be

    [(1, 2), (3, 4), (5, 6), (7, 8)]

    (Note that we insist each element only appears in a pairing once; the
    following is not a pairing:

    [(1, 1), (2, 2), (3, 4), (5, 6), (7, 8)]

    This function generates a set of len(labels)-1 such pairings of the input
    list labels, such that each element in labels is paired with each other
    element in at least one pairing

    Args:
        labels (list): list of elements

    Yields:
        pairings (list): list of pairings of elements of labels
    """
    if not labels:
        return
    if len(labels) == 1:
        yield (labels[0],)
        return

    # Determine fragment size
    fragment_size = len(labels) // 2
    frag1 = labels[:fragment_size]
    frag2 = labels[fragment_size:]

    for pairing in pair_between(frag1, frag2, len(frag2) % 2):
        yield pairing

    if len(labels) % 4 == 1:
        frag1.append(None)

    for (pairing1, pairing2) in zip(pair_within(frag1), pair_within(frag2)):

        if len(labels) % 4 == 1:
            if pairing1[-1] is None:
                yield pairing1[:-1] + pairing2
            else:
                extra_pair = ((pairing1[-1], pairing2[-1]),)
                zero_index, = [
                    pair[0] for pair in pairing1[:-1] if pair[1] is None
                ]
                pairing1 = tuple(
                    pair for pair in pairing1[:-1] if pair[1] is not None)
                yield pairing1 + pairing2[:-1] + extra_pair + (zero_index,)

        elif len(labels) % 4 == 2:
            extra_pair = ((pairing1[-1], pairing2[-1]),)
            yield pairing1[:-1] + pairing2[:-1] + extra_pair

        elif len(labels) % 4 == 3:
            yield pairing1[:-1] + pairing2 + (pairing1[-1],)

        else:
            yield pairing1 + pairing2


def pair_between(frag1: list, frag2: list, start_offset: int = 0) -> tuple:
    """Pairs between two fragments of a larger list

    A pairing of a list is a set of pairs of list elements. E.g. a pairing of

    labels = [1, 2, 3, 4, 5, 6, 7, 8]

    could be

    [(1, 2), (3, 4), (5, 6), (7, 8)]

    (Note that we insist each element only appears in a pairing once; the
    following is not a pairing:

    [(1, 1), (2, 2), (3, 4), (5, 6), (7, 8)]

    This function generates a set of pairings between elements of frag1
    and frag2 such that element1 in frag 1 and element2 in frag2,
    the pair (element1, element2) is found exactly once within the pairing.

    Args:
        frag1, frag2 (lists): the elements to be paired
        start_offset (int): prevents the first start_offset pairings
            from being yielded

    Yields:
        pairing tuple: the desired pairings, followed by
            any unpaired elements
    """

    num_iter = max(len(frag1), len(frag2))
    num_pairs = min(len(frag1), len(frag2))

    for index_offset in range(start_offset, num_iter):

        if len(frag1) > len(frag2):
            pairing = tuple(
                (frag1[(index + index_offset) % len(frag1)], frag2[index])
                for index in range(num_pairs))
            pairing += tuple(frag1[index % len(frag1)] for index in range(
                len(frag2) + index_offset,
                len(frag1) + index_offset))
        else:
            pairing = tuple(
                (frag1[index], frag2[(index + index_offset) % len(frag2)])
                for index in range(num_pairs))
        if len(frag2) > len(frag1):
            pairing += tuple(frag2[index % len(frag2)] for index in range(
                len(frag1) + index_offset,
                len(frag2) + index_offset))

        yield pairing


def _loop_iterator(func, *params):
    generator = func(*params)
    looped = False
    num_loops = 0
    while True:
        for res in generator:
            yield res, looped
        looped = True
        num_loops += 1
        if num_loops > MAX_LOOPS:
            raise ValueError(
                'Number of loops exceeded maximum allowed.')  # pragma: no cover
        generator = func(*params)


def _gen_partitions(labels, min_size=4):
    """
    Generates a set of exponentially smaller partitions of a set

    Args:
        labels(list): list to be partitioned
    """
    if len(labels) == 1:
        yield (labels,)
        return
    partitions = (labels[:len(labels) // 2], labels[len(labels) // 2:])
    while True:
        yield partitions
        if len(partitions[-1]) < min_size:
            return
        new_partitions = []
        for part in partitions:
            new_partitions.append(part[:len(part) // 2])
            new_partitions.append(part[len(part) // 2:])
        partitions = new_partitions


def _gen_pairings_between_partitions(parta, partb):
    if len(parta + partb) < 5:
        yield (tuple(parta), tuple(partb))
    splita = [parta[:len(parta) // 2], parta[len(parta) // 2:]]
    splitb = [partb[:len(partb) // 2], partb[len(partb) // 2:]]
    for a, b in ((0, 0), (0, 1), (1, 0), (1, 1)):
        if max(len(splita[a]), len(splitb[b])) < 2:
            continue
        if min(len(splita[1 - a]), len(splitb[1 - b])) < 1:
            continue  # pragma: no cover
        gen_a = _loop_iterator(pair_within, splita[a])
        gen_b = _loop_iterator(pair_within, splitb[b])
        num_iter = max(
            len(splitb[b]) - 1 + len(splitb[b]) % 2,
            len(splita[a]) - 1 + len(splita[a]) % 2)
        for _ in range(num_iter):
            pair_a, _ = next(gen_a)
            pair_b, _ = next(gen_b)
            gen_ab = pair_between(splita[1 - a], splitb[1 - b])
            for pair_ab in gen_ab:
                yield pair_a + pair_b + pair_ab


def pair_within_simultaneously(labels: list) -> tuple:
    """Generates simultaneous pairings between four-element combinations

    A pairing of a list is a set of pairs of list elements. E.g. a pairing of

    labels = [1, 2, 3, 4, 5, 6, 7, 8]

    could be

    [(1, 2), (3, 4), (5, 6), (7, 8)]

    (Note that we insist each element only appears in a pairing once; the
    following is not a pairing:

    [(1, 1), (2, 2), (3, 4), (5, 6), (7, 8)]

    This function generates a set of pairings such that for every four elements
    (i,j,k,l) in 'labels', there exists one pairing containing both (i,j) and
    (k,l)

    Args:
        labels(list): list of elements to be paired

    Yields:
        pairings(tuple of pairs): the desired pairings
    """

    if len(labels) <= 3:
        return

    for partition in _gen_partitions(labels):
        generator_list = [
            _loop_iterator(pair_within, partition[j])
            for j in range(len(partition))
        ]
        for dummy1 in range(len(partition[-2]) - 1 + len(partition[-2]) % 2):
            pairing = tuple()
            for generator in generator_list[::2]:
                pairing = pairing + next(generator)[0]
            for dummy2 in range(
                    len(partition[-1]) - 1 + len(partition[-1]) % 2):
                pairing2 = tuple(pairing)
                for generator in generator_list[1::2]:
                    pairing2 = pairing2 + next(generator)[0]
                yield pairing2

        if len(partition[-1]) < 3:
            continue

        for partition_pairing in pair_within(partition):
            generator_list = [
                _loop_iterator(_gen_pairings_between_partitions, part_a, part_b)
                for part_a, part_b in partition_pairing
            ]
            while True:
                pairing = tuple()
                looped = True
                for generator in generator_list:
                    this_pairing, this_looped = next(generator)
                    pairing += this_pairing
                    if this_looped is False:
                        looped = False
                if looped is True:
                    break
                yield pairing


def _get_padding(num_bins, bin_size):
    """
    For parallel iteration: gets the smallest number L' >= bin_size
    such that num_bins is smaller than the lowest factor of L'.
    """
    trial_size = bin_size
    while True:
        success_flag = True
        for divisor in range(2, num_bins - 1):
            if trial_size % divisor == 0:
                success_flag = False
                break
        if success_flag:
            return trial_size
        trial_size += 1


def _asynchronous_iter(iterators, flatten=False):
    """
    Iterates over a set of K iterators with max L elements to
    generate all pairs between them in O(L^2 + 2L log(L) + log(L)^2),
    assuming L>>K. When appropriate, calls a different iterator
    optimized for small lists.

    Args:
        iterators(list of iterators): iterators to be passed
        flatten(boolean): whether to concatenate or join the results.
    Yields:
        next_result(list of results): the joined/concatenated set of results.
    """
    iterator_lists = [list(iterator) for iterator in iterators]
    num_lists = len(iterator_lists)
    list_size = max([len(lst) for lst in iterator_lists])

    # Edge cases
    if list_size == 1:
        next_res = [
            iterator[0] if iterator else None for iterator in iterator_lists
        ]
        if flatten:
            next_res = [x for result in next_res if result for x in result]
        yield tuple(next_res)
        return
    if numpy.log2(num_lists + 1) * list_size**2 < num_lists**2:
        for next_res in _asynchronous_iter_small_lists(iterator_lists, flatten):
            yield next_res
        return

    new_size = _get_padding(num_lists, list_size)
    for lst in iterator_lists:
        lst += [None] * (new_size - len(lst))

    for j in range(new_size):
        for l in range(new_size):
            next_res = [
                iterator_lists[k][(j * k + l) % new_size]
                for k in range(num_lists - 1)
            ]
            next_res.append(iterator_lists[-1][j])
            if flatten:
                next_res = [x for result in next_res if result for x in result]
            yield tuple(next_res)


def _asynchronous_iter_small_lists(iterator_lists, flatten=False):
    """
    Iterates over a set of K iterators of max L items to generate all pairs
    between them in O(log(K)L^2) time - this is suboptimal when L>>K,
    but does not require list padding, making it better for small L.

    Args:
        iterators(list of iterators): iterators to be passed
        flatten(boolean): whether to concatenate or join the results.
    Yields:
        next_result(list of results): the joined/concatenated set of results.
    """
    for partitions in partition_iterator(iterator_lists, 2):
        for res in _asynchronous_iter(
            [_parallel_iter(partition, flatten) for partition in partitions],
                flatten):
            yield res


def _parallel_iter(iterators, flatten=False):
    """
    Iterates in parallel over a set of iterators.
    Stopped iterators are removed, so the position of any
    result is not conserved.

    Args:
        iterators(list of iterables): iterators to be passed
        flatten(boolean): whether to concatenate or join the results.
    Yields:
        next_result(list of results): the joined/concatenated set of results.
    """
    iterators = [iter(iterator) for iterator in iterators]

    while iterators:
        next_result = []
        for j in range(len(iterators) - 1, -1, -1):
            temp = next(iterators[j], None)
            if temp is None:
                del iterators[j]
            else:
                if flatten:
                    next_result = list(temp) + next_result
                else:
                    next_result = [temp] + next_result
        if next_result:
            yield tuple(next_result)


def pair_within_simultaneously_binned(binned_majoranas: list) -> tuple:
    """Generates symmetry-respecting pairings between four-elements in a list

    A pairing of a list is a set of pairs of list elements. E.g. a pairing of

    labels = [1, 2, 3, 4, 5, 6, 7, 8]

    could be

    [(1, 2), (3, 4), (5, 6), (7, 8)]

    (Note that we insist each element only appears in a pairing once; the
    following is not a pairing:

    [(1, 1), (2, 2), (3, 4), (5, 6), (7, 8)]

    This function generates a pairing of a list of Majoranas that covers
    all 2-RDM elements that conserve a set of symmetry conditions. That is,
    this function guarantees that for any four elements (i,j,k,l) in the input,
    if the corresponding RDM term gamma_igamma_jgamma_kgamma_l satisfies a
    symmetry of the system, the pairs (i,j) and (k,l) will appear simultaneously
    in at least one pairing.

    The constraints are defined by a binning of the Majoranas into bins such
    that Majoranas in bin n commute with symmetry S_i if the ith binary digit
    of n is 0.

    Args:
        binned_majoranas(list of lists of integers): majoranas to be paired,
            separated up by their symmetry bins.

    Yields:
        pairing(tuple): one of the desired pairings
    """

    # Generate all four-fold pairings within bins
    iterators = [pair_within_simultaneously(bn) for bn in binned_majoranas]
    for pairing in _parallel_iter(iterators, flatten=True):
        yield pairing

    # Iterate over all pairs within bins
    num_bins = len(binned_majoranas)
    if max([len(bn) for bn in binned_majoranas]) > 1 and num_bins > 1:
        iterators = [pair_within(bn) for bn in binned_majoranas]
        for pairing in _asynchronous_iter(iterators, flatten=True):
            yield pairing

    # Pair between bins
    for bin_gap in range(1, num_bins // 2):
        iterators = []
        for bin_index in range(num_bins):
            if bin_index < bin_index ^ bin_gap:
                iterators.append(
                    pair_between(binned_majoranas[bin_index],
                                 binned_majoranas[bin_index ^ bin_gap]))
        for pairing in _asynchronous_iter(iterators, flatten=True):
            yield pairing


def pair_within_simultaneously_symmetric(num_fermions: int,
                                         num_symmetries: int) -> tuple:
    """Generates symmetry-respecting pairings between four-elements in a list

    A pairing of a list is a set of pairs of list elements. E.g. a pairing of

    labels = [1, 2, 3, 4, 5, 6, 7, 8]

    could be

    [(1, 2), (3, 4), (5, 6), (7, 8)]

    (Note that we insist each element only appears in a pairing once; the
    following is not a pairing:

    [(1, 1), (2, 2), (3, 4), (5, 6), (7, 8)]

    This function generates a pairing of a list of Majoranas that covers
    all 2-RDM elements that conserve a set of symmetry conditions. That is,
    this function guarantees that for any four elements (i,j,k,l) in the input,
    if the corresponding RDM term gamma_igamma_jgamma_kgamma_l satisfies a
    symmetry of the system, the pairs (i,j) and (k,l) will appear simultaneously
    in at least one pairing.

    We assume in this function that each symmetry divides the set of Majoranas
    in two, indexed by their binary digits.

    Args:
        num_fermions (int) : The number of fermions to be considered (the
            number of Majoranas generated will be twice this size)
        num_symmetries (int): the number of symmetries to be respectd.
    """
    binned_majoranas = [[
        index
        for index in range(2 * num_fermions)
        if index % 2**num_symmetries == bin_index
    ]
                        for bin_index in range(2**num_symmetries)]

    for pairing in pair_within_simultaneously_binned(binned_majoranas):
        yield pairing



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests  fermion_operator.py."""
import unittest

from openfermion.ops.operators.fermion_operator import FermionOperator
from openfermion.hamiltonians import number_operator


class FermionOperatorTest(unittest.TestCase):

    def test_is_normal_ordered_empty(self):
        op = FermionOperator() * 2
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_number(self):
        op = FermionOperator('2^ 2') * -1j
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_reversed(self):
        self.assertFalse(FermionOperator('2 2^').is_normal_ordered())

    def test_is_normal_ordered_create(self):
        self.assertTrue(FermionOperator('11^').is_normal_ordered())

    def test_is_normal_ordered_annihilate(self):
        self.assertTrue(FermionOperator('0').is_normal_ordered())

    def test_is_normal_ordered_long_not(self):
        self.assertFalse(FermionOperator('0 5^ 3^ 2^ 1^').is_normal_ordered())

    def test_is_normal_ordered_outoforder(self):
        self.assertFalse(FermionOperator('0 1').is_normal_ordered())

    def test_is_normal_ordered_long_descending(self):
        self.assertTrue(FermionOperator('5^ 3^ 2^ 1^ 0').is_normal_ordered())

    def test_is_normal_ordered_multi(self):
        op = FermionOperator('4 3 2^ 2') + FermionOperator('1 2')
        self.assertFalse(op.is_normal_ordered())

    def test_is_normal_ordered_multiorder(self):
        op = FermionOperator('4 3 2 1') + FermionOperator('3 2')
        self.assertTrue(op.is_normal_ordered())

    def test_is_two_body_number_conserving_FermionOperator(self):
        op = FermionOperator()
        self.assertTrue(op.is_two_body_number_conserving())

    def test_is_two_body_number_conserving_number(self):
        op = number_operator(5, 3)
        self.assertTrue(op.is_two_body_number_conserving())

    def test_is_two_body_number_conserving_updown(self):
        op = FermionOperator(((2, 1), (4, 0)))
        self.assertTrue(op.is_two_body_number_conserving())

    def test_is_two_body_number_conserving_downup(self):
        op = FermionOperator(((2, 0), (4, 1)))
        self.assertTrue(op.is_two_body_number_conserving())

    def test_is_two_body_number_conserving_downup_badspin(self):
        op = FermionOperator(((2, 0), (3, 1)))
        self.assertFalse(op.is_two_body_number_conserving(True))

    def test_is_two_body_number_conserving_three(self):
        op = FermionOperator(((0, 1), (2, 1), (4, 0)))
        self.assertFalse(op.is_two_body_number_conserving())

    def test_is_two_body_number_conserving_out_of_order(self):
        op = FermionOperator(((0, 1), (2, 0), (1, 1), (3, 0)))
        self.assertTrue(op.is_two_body_number_conserving())



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""QubitOperator stores a sum of Pauli operators acting on qubits."""

import numpy

from openfermion.ops.operators.symbolic_operator import SymbolicOperator

# Define products of all Pauli operators for symbolic multiplication.
_PAULI_OPERATOR_PRODUCTS = {
    ('I', 'I'): (1., 'I'),
    ('I', 'X'): (1., 'X'),
    ('X', 'I'): (1., 'X'),
    ('I', 'Y'): (1., 'Y'),
    ('Y', 'I'): (1., 'Y'),
    ('I', 'Z'): (1., 'Z'),
    ('Z', 'I'): (1., 'Z'),
    ('X', 'X'): (1., 'I'),
    ('Y', 'Y'): (1., 'I'),
    ('Z', 'Z'): (1., 'I'),
    ('X', 'Y'): (1.j, 'Z'),
    ('X', 'Z'): (-1.j, 'Y'),
    ('Y', 'X'): (-1.j, 'Z'),
    ('Y', 'Z'): (1.j, 'X'),
    ('Z', 'X'): (1.j, 'Y'),
    ('Z', 'Y'): (-1.j, 'X')
}


class QubitOperator(SymbolicOperator):
    """
    A sum of terms acting on qubits, e.g., 0.5 * 'X0 X5' + 0.3 * 'Z1 Z2'.

    A term is an operator acting on n qubits and can be represented as:

    coefficient * local_operator[0] x ... x local_operator[n-1]

    where x is the tensor product. A local operator is a Pauli operator
    ('I', 'X', 'Y', or 'Z') which acts on one qubit. In math notation a term
    is, for example, 0.5 * 'X0 X5', which means that a Pauli X operator acts
    on qubit 0 and 5, while the identity operator acts on all other qubits.

    A QubitOperator represents a sum of terms acting on qubits and overloads
    operations for easy manipulation of these objects by the user.

    Note for a QubitOperator to be a Hamiltonian which is a hermitian
    operator, the coefficients of all terms must be real.

    .. code-block:: python

        hamiltonian = 0.5 * QubitOperator('X0 X5') + 0.3 * QubitOperator('Z0')

    QubitOperator is a subclass of SymbolicOperator. Importantly, it has
    attributes set as follows::

        actions = ('X', 'Y', 'Z')
        action_strings = ('X', 'Y', 'Z')
        action_before_index = True
        different_indices_commute = True

    See the documentation of SymbolicOperator for more details.

    Example:
        .. code-block:: python

            ham = ((QubitOperator('X0 Y3', 0.5)
                    + 0.6 * QubitOperator('X0 Y3')))
            # Equivalently
            ham2 = QubitOperator('X0 Y3', 0.5)
            ham2 += 0.6 * QubitOperator('X0 Y3')

    Note:
        Adding QubitOperators is faster using += (as this
        is done by in-place addition). Specifying the coefficient
        during initialization is faster than multiplying a QubitOperator
        with a scalar.
    """

    @property
    def actions(self):
        """The allowed actions."""
        return ('X', 'Y', 'Z')

    @property
    def action_strings(self):
        """The string representations of the allowed actions."""
        return ('X', 'Y', 'Z')

    @property
    def action_before_index(self):
        """Whether action comes before index in string representations."""
        return True

    @property
    def different_indices_commute(self):
        """Whether factors acting on different indices commute."""
        return True

    def renormalize(self):
        """Fix the trace norm of an operator to 1"""
        norm = self.induced_norm(2)
        if numpy.isclose(norm, 0.0):
            raise ZeroDivisionError('Cannot renormalize empty or zero operator')
        else:
            self /= norm

    def _simplify(self, term, coefficient=1.0):
        """Simplify a term using commutator and anti-commutator relations."""
        if not term:
            return coefficient, term

        term = sorted(term, key=lambda factor: factor[0])

        new_term = []
        left_factor = term[0]
        for right_factor in term[1:]:
            left_index, left_action = left_factor
            right_index, right_action = right_factor

            # Still on the same qubit, keep simplifying.
            if left_index == right_index:
                new_coefficient, new_action = _PAULI_OPERATOR_PRODUCTS[
                    left_action, right_action]
                left_factor = (left_index, new_action)
                coefficient *= new_coefficient

            # Reached different qubit, save result and re-initialize.
            else:
                if left_action != 'I':
                    new_term.append(left_factor)
                left_factor = right_factor

        # Save result of final iteration.
        if left_factor[1] != 'I':
            new_term.append(left_factor)

        return coefficient, tuple(new_term)



from itertools import product
import os
import numpy as np
from openfermion.config import DATA_DIRECTORY
from openfermion.chem import MolecularData, make_reduced_hamiltonian
from openfermion.linalg import wedge
from openfermion.linalg.erpa import singlet_erpa, erpa_eom_hamiltonian
from openfermion.transforms.opconversions import get_fermion_operator
from openfermion.ops.representations import InteractionRDM
from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import normal_ordered
from openfermion.transforms.repconversions import get_interaction_operator
from openfermion.utils.commutators import commutator


def test_h2_rpa():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)
    hf_opdm = np.diag([1] * molecule.n_electrons + [0] *
                      (molecule.n_qubits - molecule.n_electrons))
    hf_tpdm = 2 * wedge(hf_opdm, hf_opdm, (1, 1), (1, 1))

    pos_spectrum, xy_eigvects, basis = singlet_erpa(hf_tpdm,
                                                    reduced_ham.two_body_tensor)
    assert np.isclose(pos_spectrum, 0.92926444)  # pyscf-rpa value
    assert isinstance(xy_eigvects, np.ndarray)
    assert isinstance(basis, dict)


def test_erpa_eom_ham_h2():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)
    rha_fermion = get_fermion_operator(reduced_ham)
    permuted_hijkl = np.einsum('ijlk', reduced_ham.two_body_tensor)
    opdm = np.diag([1] * molecule.n_electrons + [0] *
                   (molecule.n_qubits - molecule.n_electrons))
    tpdm = 2 * wedge(opdm, opdm, (1, 1), (1, 1))
    rdms = InteractionRDM(opdm, tpdm)
    dim = reduced_ham.one_body_tensor.shape[0] // 2
    full_basis = {}  # erpa basis.  A, B basis in RPA language
    cnt = 0
    for p, q in product(range(dim), repeat=2):
        if p < q:
            full_basis[(p, q)] = cnt
            full_basis[(q, p)] = cnt + dim * (dim - 1) // 2
            cnt += 1
    for rkey in full_basis.keys():
        p, q = rkey
        for ckey in full_basis.keys():
            r, s = ckey
            for sigma, tau in product([0, 1], repeat=2):
                test = erpa_eom_hamiltonian(permuted_hijkl, tpdm, 2 * q + sigma,
                                            2 * p + sigma, 2 * r + tau,
                                            2 * s + tau).real
                qp_op = FermionOperator(
                    ((2 * q + sigma, 1), (2 * p + sigma, 0)))
                rs_op = FermionOperator(((2 * r + tau, 1), (2 * s + tau, 0)))
                erpa_op = normal_ordered(
                    commutator(qp_op, commutator(rha_fermion, rs_op)))
                true = rdms.expectation(get_interaction_operator(erpa_op))
                assert np.isclose(true, test)


def test_erpa_eom_ham_lih():
    filename = os.path.join(DATA_DIRECTORY, "H1-Li1_sto-3g_singlet_1.45.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)
    rha_fermion = get_fermion_operator(reduced_ham)
    permuted_hijkl = np.einsum('ijlk', reduced_ham.two_body_tensor)
    opdm = np.diag([1] * molecule.n_electrons + [0] *
                   (molecule.n_qubits - molecule.n_electrons))
    tpdm = 2 * wedge(opdm, opdm, (1, 1), (1, 1))
    rdms = InteractionRDM(opdm, tpdm)
    dim = 3  # so we don't do the full basis.  This would make the test long
    full_basis = {}  # erpa basis.  A, B basis in RPA language
    cnt = 0
    # start from 1 to make test shorter
    for p, q in product(range(1, dim), repeat=2):
        if p < q:
            full_basis[(p, q)] = cnt
            full_basis[(q, p)] = cnt + dim * (dim - 1) // 2
            cnt += 1
    for rkey in full_basis.keys():
        p, q = rkey
        for ckey in full_basis.keys():
            r, s = ckey
            for sigma, tau in product([0, 1], repeat=2):
                test = erpa_eom_hamiltonian(permuted_hijkl, tpdm, 2 * q + sigma,
                                            2 * p + sigma, 2 * r + tau,
                                            2 * s + tau).real
                qp_op = FermionOperator(
                    ((2 * q + sigma, 1), (2 * p + sigma, 0)))
                rs_op = FermionOperator(((2 * r + tau, 1), (2 * s + tau, 0)))
                erpa_op = normal_ordered(
                    commutator(qp_op, commutator(rha_fermion, rs_op)))
                true = rdms.expectation(get_interaction_operator(erpa_op))
                assert np.isclose(true, test)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import unittest

import numpy

from openfermion.transforms.opconversions import get_fermion_operator
from openfermion.ops.representations import DiagonalCoulombHamiltonian
from openfermion.testing.testing_utils import \
    random_diagonal_coulomb_hamiltonian


class DiagonalCoulombHamiltonianTest(unittest.TestCase):

    def test_init(self):
        one_body = numpy.array([[2., 3.], [3., 5.]])
        two_body = numpy.array([[7., 11.], [11., 13.]])
        constant = 17.
        op = DiagonalCoulombHamiltonian(one_body, two_body, constant)
        self.assertTrue(
            numpy.allclose(op.one_body, numpy.array([[9., 3.], [3., 18.]])))
        self.assertTrue(
            numpy.allclose(op.two_body, numpy.array([[0., 11.], [11., 0.]])))
        self.assertAlmostEqual(op.constant, 17.)

    def test_multiply(self):
        n_qubits = 5
        op1 = random_diagonal_coulomb_hamiltonian(n_qubits)
        op2 = op1 * 1.5
        op3 = 1.5 * op1
        self.assertEqual(
            get_fermion_operator(op1) * 1.5, get_fermion_operator(op2),
            get_fermion_operator(op3))

    def test_divide(self):
        n_qubits = 5
        op1 = random_diagonal_coulomb_hamiltonian(n_qubits)
        op2 = op1 / 1.5
        self.assertEqual(
            get_fermion_operator(op1) / 1.5, get_fermion_operator(op2))

    def test_exceptions(self):
        mat1 = numpy.array([[2., 3.], [3., 5.]])
        mat2 = numpy.array([[2., 3. + 1.j], [3 - 1.j, 5.]])
        mat3 = numpy.array([[2., 3.], [4., 5.]])
        with self.assertRaises(ValueError):
            _ = DiagonalCoulombHamiltonian(mat1, mat2)
        with self.assertRaises(ValueError):
            _ = DiagonalCoulombHamiltonian(mat2, mat3)
        with self.assertRaises(ValueError):
            _ = DiagonalCoulombHamiltonian(mat3, mat1)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Defines index mappings."""
import numpy


def up_index(index):
    """Function to return up-orbital index given a spatial orbital index.

    Args:
        index (int): spatial orbital index

    Returns:
        An integer representing the index of the associated spin-up orbital
    """
    return 2 * index


def down_index(index):
    """Function to return down-orbital index given a spatial orbital index.

    Args:
        index (int): spatial orbital index

    Returns:
        An integer representing the index of the associated spin-down orbital
    """
    return 2 * index + 1


def up_then_down(mode_idx, num_modes):
    """ up then down reordering, given the operator has the default even-odd
     ordering. Otherwise this function will reorder indices where all even
     indices now come before odd indices.

     Example:
         0,1,2,3,4,5 -> 0,2,4,1,3,5

    The function takes in the index of the mode that will be relabeled and
    the total number modes.

    Args:
        mode_idx (int): the mode index that is being reordered
        num_modes (int): the total number of modes of the operator.

    Returns (int): reordered index of the mode.
    """
    halfway = int(numpy.ceil(num_modes / 2.))

    if mode_idx % 2 == 0:
        return mode_idx // 2

    return mode_idx // 2 + halfway



# BSD 3-Clause License
#
# Copyright (c) 2018 Rigetti & Co, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# pylint: disable=C
from itertools import product
import numpy as np
import pytest
from openfermion.third_party.representability._higham import (
    heaviside, higham_polynomial, higham_root, map_to_tensor, map_to_matrix,
    fixed_trace_positive_projection)


def test_heaviside():
    assert np.isclose(heaviside(0), 1.0)
    assert np.isclose(heaviside(0.5), 1.0)
    assert np.isclose(heaviside(-0.5), 0.0)
    assert np.isclose(heaviside(-0.5, -1), 1.0)
    assert np.isclose(heaviside(-2, -1), 0)


def test_highham_polynomial():
    eigs = np.arange(10)
    assert np.isclose(higham_polynomial(eigs, eigs[-1]), 0.0)
    assert np.isclose(higham_polynomial(eigs, 0), sum(eigs))
    assert np.isclose(higham_polynomial(eigs, 5), sum(eigs[5:] - 5))
    assert np.isclose(higham_polynomial(eigs, 8), sum(eigs[8:] - 8))


def test_higham_root():
    dim = 20
    np.random.seed(42)
    mat = np.random.random((dim, dim))
    mat = 0.5 * (mat + mat.T)
    w, _ = np.linalg.eigh(mat)
    target_trace = np.round(w[-1] - 1)
    sigma = higham_root(w, target_trace)
    assert np.isclose(higham_polynomial(w, shift=sigma), target_trace)

    with pytest.raises(ValueError):
        higham_root(w, target_trace=-1)

    tw = higham_root(w, target_trace=0)
    assert np.isclose(tw, w[-1])


def test_matrix_2_tensor():
    dim = 10
    np.random.seed(42)
    mat = np.random.random((dim**2, dim**2))
    mat = 0.5 * (mat + mat.T)
    tensor = map_to_tensor(mat)
    for p, q, r, s in product(range(dim), repeat=4):
        assert np.isclose(tensor[p, q, r, s], mat[p * dim + q, r * dim + s])

    test_mat = map_to_matrix(tensor)
    assert np.allclose(test_mat, mat)

    with pytest.raises(TypeError):
        map_to_tensor(np.zeros((4, 4, 4, 4)))

    with pytest.raises(TypeError):
        map_to_matrix(np.zeros((4, 4)))


def test_reconstruction():
    dim = 20
    np.random.seed(42)
    mat = np.random.random((dim, dim))
    mat = 0.5 * (mat + mat.T)
    test_mat = np.zeros_like(mat)
    w, v = np.linalg.eigh(mat)
    for i in range(w.shape[0]):
        test_mat += w[i] * v[:, [i]].dot(v[:, [i]].T)
    assert np.allclose(test_mat - mat, 0.0)

    test_mat = fixed_trace_positive_projection(mat, np.trace(mat))
    assert np.isclose(np.trace(test_mat), np.trace(mat))
    w, v = np.linalg.eigh(test_mat)
    assert np.all(w >= -(float(4.0E-15)))

    mat = np.arange(16).reshape((4, 4))
    mat = 0.5 * (mat + mat.T)
    mat_tensor = map_to_tensor(mat)
    trace_mat = np.trace(mat)
    true_mat = fixed_trace_positive_projection(mat, trace_mat)
    test_mat = map_to_matrix(
        fixed_trace_positive_projection(mat_tensor, trace_mat))
    assert np.allclose(true_mat, test_mat)

    assert np.allclose(true_mat,
                       fixed_trace_positive_projection(true_mat, trace_mat))


def test_mlme():
    """
    Test from fig 1 of maximum likelihood minimum effort!
    """
    eigs = np.array(
        list(reversed([3.0 / 5, 1.0 / 2, 7.0 / 20, 1.0 / 10, -11.0 / 20])))
    target_trace = 1.0
    sigma = higham_root(eigs, target_trace)
    shifted_eigs = np.multiply(heaviside(eigs - sigma), (eigs - sigma))
    assert np.allclose(shifted_eigs, [0, 0, 1.0 / 5, 7.0 / 20, 9.0 / 20])



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""IsingOperator stores an Ising-type Hamiltonian, i.e. a sum of products of
Zs."""

from collections import defaultdict

from openfermion.ops.operators.symbolic_operator import SymbolicOperator


class IsingOperator(SymbolicOperator):
    """The IsingOperator class provides an analytic representation of an
    Ising-type Hamiltonian, i.e. a sum of product of Zs.

    IsingOperator is a subclass of SymbolicOperator. Importantly, it has
    attributes set as follows:

        actions = ('Z')
        action_strings = ('Z')
        action_before_index = True
        different_indices_commute = True

    See the documentation of SymbolicOperator for more details.
    """

    @property
    def actions(self):
        """The allowed actions."""
        return ('Z',)

    @property
    def action_strings(self):
        """The string representations of the allowed actions."""
        return ('Z',)

    @property
    def action_before_index(self):
        """Whether action comes before index in string representations."""
        return True

    @property
    def different_indices_commute(self):
        """Whether factors acting on different indices commute."""
        return True

    def _simplify(self, term, coefficient=1.0):
        powers = defaultdict(int)
        for factor in term:
            powers[factor[0]] += 1
        odd_powers = sorted(i for i, p in powers.items() if p % 2)
        new_term = tuple((i, 'Z') for i in odd_powers)
        return coefficient, new_term



# BSD 3-Clause License
#
# Copyright (c) 2018 Rigetti & Co, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
This module contains methods to find the closest positive semidefinite matrix
with fixed trace by the method in arXiv 1707.01022v1 and
 N.J. Higham, Linear Algebra and Its Applications 103, 103 (1998)
"""
from itertools import product
import numpy as np


@np.vectorize
def heaviside(x, bias=0) -> int:
    """
    Heaviside function Theta(x - bias)

    returns 1 if x >= bias else 0

    :param x: floating point number as input to heavisde
    :param bias: shift on the heaviside function
    :return: 1 or 0 int
    """
    indicator = 1 if x >= bias else 0
    return indicator


def higham_polynomial(eigenvalues, shift):
    """
    Calculate the higham_polynomial

    Args:
        eigenvalues:  vector of eigenvalues
        shift: where to put the bias for the heaviside function
    """
    heaviside_indicator = np.asarray(heaviside(eigenvalues, bias=shift))
    return heaviside_indicator.T.dot(eigenvalues - shift)


def higham_root(eigenvalues, target_trace, epsilon=1.0E-15):
    """
    Find the root of f(sigma) = sum_{j}Theta(l_{i} - sigma)(l_{i} - sigma) = T

    Args:
        eigenvalues: ordered list of eigenvalues from least to greatest
        target_trace: trace to maintain on new matrix
        epsilon: precision on bisection linesearch
    """
    if target_trace < 0.0:
        raise ValueError("Target trace needs to be a non-negative number")

    # when we want the trace to be zero
    if np.isclose(target_trace, 0.0):
        return eigenvalues[-1]

    # find top sigma
    sigma = eigenvalues[-1]
    while higham_polynomial(eigenvalues, sigma) < target_trace:
        sigma -= eigenvalues[-1]

    sigma_low = sigma
    sigma_high = eigenvalues[-1]

    while sigma_high - sigma_low >= epsilon:
        midpoint = sigma_high - (sigma_high - sigma_low) / 2.0
        if higham_polynomial(eigenvalues, midpoint) < target_trace:
            sigma_high = midpoint
        else:
            sigma_low = midpoint

    return sigma_high


def map_to_matrix(mat):
    if mat.ndim != 4:
        raise TypeError(
            "I only map rank-4 tensors to matices with symmetric support")
    dim = mat.shape[0]
    matform = np.zeros((dim**2, dim**2))
    for p, q, r, s in product(range(dim), repeat=4):
        assert np.isclose(mat[p, q, r, s].imag, 0.0)
        matform[p * dim + q, r * dim + s] = mat[p, q, r, s].real
    return matform


def map_to_tensor(mat):
    if mat.ndim != 2:
        raise TypeError(
            "I only map matrices to rank-4 tensors with symmetric support")
    dim = int(np.sqrt(mat.shape[0]))
    tensor_form = np.zeros((dim, dim, dim, dim))
    for p, q, r, s in product(range(dim), repeat=4):
        tensor_form[p, q, r, s] = mat[p * dim + q, r * dim + s]
    return tensor_form


def fixed_trace_positive_projection(bmat, target_trace):
    """
    Perform the positive projection with fixed trace

    Args:
        bmat:  Symmetric matrix to perform positive projection on
        target_trace:  What the trace should be
    Returns: new matrix that has the target trace and is positive semidefinite
    """
    bmat = np.asarray(bmat)
    map_to_four_tensor = False
    if bmat.ndim == 4:
        bmat = map_to_matrix(bmat)
        map_to_four_tensor = True

    # symmeterize bmat
    if np.allclose(bmat - bmat.conj().T, np.zeros_like(bmat)):
        bmat = 0.5 * (bmat + bmat.conj().T)

    w, v = np.linalg.eigh(bmat)
    if np.all(w >= -1.0 * float(1.0E-14)) and np.isclose(
            np.sum(w), target_trace):
        purified_matrix = bmat
    else:
        sigma = higham_root(w, target_trace)
        shifted_eigs = np.multiply(heaviside(w - sigma), (w - sigma))
        purified_matrix = np.zeros_like(bmat)
        for i in range(w.shape[0]):
            purified_matrix += shifted_eigs[i] * \
                               v[:, [i]].dot(v[:, [i]].conj().T)

    if map_to_four_tensor:
        purified_matrix = map_to_tensor(purified_matrix)

    return purified_matrix



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for qubit_partitioning.py"""
import unittest

from openfermion.ops.operators import (QubitOperator, FermionOperator,
                                       BosonOperator)

from .qubit_partitioning import (binary_partition_iterator, partition_iterator,
                                 pauli_string_iterator,
                                 group_into_tensor_product_basis_sets)


class BinaryPartitionIteratorTest(unittest.TestCase):

    def test_num_partitions(self):
        qubit_list = range(6)
        bpi = binary_partition_iterator(qubit_list)
        count = 0
        for _, _ in bpi:
            count += 1
        self.assertEqual(count, 3)

    def test_partition_size(self):
        qubit_list = range(6)
        bpi = binary_partition_iterator(qubit_list)
        for p1, p2 in bpi:
            self.assertEqual(len(p1), 3)
            self.assertEqual(len(p2), 3)

    def test_partition_size_odd(self):
        qubit_list = range(7)
        bpi = binary_partition_iterator(qubit_list)
        for p1, p2 in bpi:
            self.assertEqual(len(p1), 4)
            self.assertEqual(len(p2), 3)

    def test_partitioning(self):
        qubit_list = list(range(6))
        for i in range(6):
            for j in range(i + 1, 6):
                print(i, j)
                flag = False
                bpi = binary_partition_iterator(qubit_list)
                for partition in bpi:
                    print(type(partition))
                    self.assertTrue(type(partition) is tuple)
                    p1, p2 = partition
                    print(p1, p2)
                    if (i in p1 and j in p2) or (j in p1 and i in p2):
                        flag = True
                self.assertTrue(flag)

    def test_partitioning_odd(self):
        qubit_list = list(range(7))
        for i in range(7):
            for j in range(i + 1, 7):
                print(i, j)
                flag = False
                bpi = binary_partition_iterator(qubit_list)
                for p1, p2 in bpi:
                    print(p1, p2)
                    if (i in p1 and j in p2) or (j in p1 and i in p2):
                        flag = True
                self.assertTrue(flag)

    def test_exception(self):
        with self.assertRaises(ValueError):
            bpi = binary_partition_iterator([])
            next(bpi)

    def test_partition_of_two(self):
        bpi = binary_partition_iterator([0, 1])
        count = 0
        for p1, p2 in bpi:
            count += 1
            self.assertEqual(p1[0], 0)
            self.assertEqual(p2[0], 1)
        self.assertEqual(count, 1)

    def test_zero_counting(self):
        bpi = binary_partition_iterator([0, 1], 0)
        with self.assertRaises(StopIteration):
            next(bpi)


class PartitionIteratorTest(unittest.TestCase):

    def test_unary_case(self):
        qubit_list = list(range(6))
        bpi = partition_iterator(qubit_list, 1)
        for p1, in bpi:
            self.assertEqual(p1, qubit_list)

    def test_binary_case(self):
        qubit_list = list(range(6))
        for i in range(6):
            for j in range(i + 1, 6):
                print(i, j)
                flag = False
                bpi = partition_iterator(qubit_list, 2)
                for p1, p2 in bpi:
                    print(p1, p2)
                    if (i in p1 and j in p2) or (j in p1 and i in p2):
                        flag = True
                self.assertTrue(flag)

    def test_exception(self):
        with self.assertRaises(ValueError):
            pi = partition_iterator([1, 2], 3)
            next(pi)

    def test_threepartition_three(self):
        bpi = partition_iterator([1, 2, 3], 3)
        count = 0
        for partition in bpi:
            print(type(partition))
            self.assertTrue(type(partition) is tuple)
            p1, p2, p3 = partition
            print(p1, p2, p3)
            self.assertEqual(len(p1), 1)
            self.assertEqual(p1[0], 1)
            self.assertEqual(len(p2), 1)
            self.assertEqual(p2[0], 2)
            self.assertEqual(len(p3), 1)
            self.assertEqual(p3[0], 3)
            count += 1
        self.assertEqual(count, 1)

    def test_partition_three(self):
        for num_qubits in range(1, 16):
            qubit_list = list(range(num_qubits))
            for i in range(num_qubits):
                for j in range(i + 1, num_qubits):
                    for k in range(j + 1, num_qubits):
                        print('Testing {}, {}, {}'.format(i, j, k))
                        pi = partition_iterator(qubit_list, 3)
                        count = 0
                        for p1, p2, p3 in pi:
                            self.assertEqual(
                                len(p1) + len(p2) + len(p3), len(qubit_list))
                            self.assertEqual(set(p1 + p2 + p3), set(qubit_list))
                            print('Partition obtained: ', p1, p2, p3)
                            if max(
                                    sum(1
                                        for x in p
                                        if x in [i, j, k])
                                    for p in [p1, p2, p3]) == 1:
                                count += 1
                        print('count = {}'.format(count))
                        self.assertTrue(count > 0)
                        print()


class PauliStringIteratorTest(unittest.TestCase):

    def test_eightpartition_three(self):
        for i1 in range(8):
            for i2 in range(i1 + 1, 8):
                for i3 in range(i2 + 1, 8):
                    for l1 in ['X', 'Y', 'Z']:
                        for l2 in ['X', 'Y', 'Z']:
                            for l3 in ['X', 'Y', 'Z']:
                                psg = pauli_string_iterator(8, 3)
                                count = 0
                                for pauli_string in psg:
                                    if (pauli_string[i1] == l1 and
                                            pauli_string[i2] == l2 and
                                            pauli_string[i3] == l3):
                                        count += 1
                                self.assertTrue(count > 0)

    def test_exceptions(self):
        with self.assertRaises(ValueError):
            psi = pauli_string_iterator(1, 2)
            next(psi)
        with self.assertRaises(ValueError):
            psi = pauli_string_iterator(3, -1)
            next(psi)

    def test_small_run_cases(self):
        for num_qubits in range(4, 20):
            for word_length in range(2, min(num_qubits, 5)):
                psi = pauli_string_iterator(num_qubits, word_length)
                for _ in psi:
                    pass


class GroupTensorProductBasisTest(unittest.TestCase):

    def test_demo_qubit_operator(self):
        for seed in [None, 0, 10000]:
            op = QubitOperator('X0 Y1', 2.) + QubitOperator('X1 Y2', 3.j)
            sub_operators = group_into_tensor_product_basis_sets(op, seed=seed)
            expected = {
                ((0, 'X'), (1, 'Y')): QubitOperator('X0 Y1', 2.),
                ((1, 'X'), (2, 'Y')): QubitOperator('X1 Y2', 3.j)
            }
            self.assertEqual(sub_operators, expected)

            op = QubitOperator('X0 Y1', 2.) + QubitOperator('Y1 Y2', 3.j)
            sub_operators = group_into_tensor_product_basis_sets(op, seed=seed)
            expected = {((0, 'X'), (1, 'Y'), (2, 'Y')): op}
            self.assertEqual(sub_operators, expected)

            op = QubitOperator('', 4.) + QubitOperator('X1', 2.j)
            sub_operators = group_into_tensor_product_basis_sets(op, seed=seed)
            expected = {((1, 'X'),): op}
            self.assertEqual(sub_operators, expected)

            op = (QubitOperator('X0 X1', 0.1) + QubitOperator('X1 X2', 2.j) +
                  QubitOperator('Y2 Z3', 3.) + QubitOperator('X3 Z4', 5.))
            sub_operators = group_into_tensor_product_basis_sets(op, seed=seed)
            expected1 = {
                ((0, 'X'), (1, 'X'), (2, 'X'), (3, 'X'), (4, 'Z')):
                (QubitOperator('X0 X1', 0.1) + QubitOperator('X1 X2', 2.j) +
                 QubitOperator('X3 Z4', 5.)),
                ((2, 'Y'), (3, 'Z')):
                QubitOperator('Y2 Z3', 3.)
            }
            expected2 = {
                ((0, 'X'), (1, 'X'), (2, 'Y'), (3, 'Z')):
                (QubitOperator('X0 X1', 0.1) + QubitOperator('Y2 Z3', 3.)),
                ((1, 'X'), (2, 'X'), (3, 'X'), (4, 'Z')):
                (QubitOperator('X1 X2', 2.j) + QubitOperator('X3 Z4', 5.))
            }
            self.assertTrue(sub_operators == expected1 or
                            sub_operators == expected2)

    def test_empty_qubit_operator(self):
        sub_operators = group_into_tensor_product_basis_sets(QubitOperator())
        self.assertTrue(sub_operators == {})

    def test_fermion_operator_bad_type(self):
        with self.assertRaises(TypeError):
            _ = group_into_tensor_product_basis_sets(FermionOperator())

    def test_boson_operator_bad_type(self):
        with self.assertRaises(TypeError):
            _ = group_into_tensor_product_basis_sets(BosonOperator())

    def test_none_bad_type(self):
        with self.assertRaises(TypeError):
            _ = group_into_tensor_product_basis_sets(None)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""tests for operator_tapering.py"""

import unittest
from openfermion.ops.operators import FermionOperator, BosonOperator
from openfermion.transforms.repconversions.operator_tapering import (
    freeze_orbitals, prune_unused_indices)


class FreezeOrbitalsTest(unittest.TestCase):

    def test_freeze_orbitals_nonvanishing(self):
        op = FermionOperator(((1, 1), (1, 0), (0, 1), (2, 0)))
        op_frozen = freeze_orbitals(op, [1])
        expected = FermionOperator(((0, 1), (1, 0)), -1)
        self.assertEqual(op_frozen, expected)

    def test_freeze_orbitals_vanishing(self):
        op = FermionOperator(((1, 1), (2, 0)))
        op_frozen = freeze_orbitals(op, [], [2])
        self.assertEqual(len(op_frozen.terms), 0)


class PruneUnusedIndicesTest(unittest.TestCase):

    def test_prune(self):
        for LadderOp in (FermionOperator, BosonOperator):
            op = LadderOp(((1, 1), (8, 1), (3, 0)), 0.5)
            op = prune_unused_indices(op)
            expected = LadderOp(((0, 1), (2, 1), (1, 0)), 0.5)
            self.assertTrue(expected == op)


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for Hubbard model lattice module."""

import itertools
import random
import pytest

from openfermion.utils import (HubbardSquareLattice, SpinPairs, Spin)


def test_spin():
    lattice = HubbardSquareLattice(3, 3)
    assert tuple(lattice.spin_indices) == (Spin.UP, Spin.DOWN)


@pytest.mark.parametrize("x_dimension,y_dimension,n_dofs,spinless,periodic",
                         itertools.product(random.sample(range(3, 10), 3),
                                           random.sample(range(3, 10), 3),
                                           range(1, 4), (False, True),
                                           (False, True)))
def test_hubbard_square_lattice(x_dimension, y_dimension, n_dofs, spinless,
                                periodic):
    lattice = HubbardSquareLattice(x_dimension,
                                   y_dimension,
                                   n_dofs=n_dofs,
                                   spinless=spinless,
                                   periodic=periodic)

    n_spin_values = 2 - spinless
    sites = tuple(
        (x, y)
        for y, x in itertools.product(range(y_dimension), range(x_dimension)))
    site_indices = tuple(lattice.to_site_index(site) for site in sites)
    assert (sites == tuple(
        lattice.from_site_index(site_index) for site_index in site_indices))
    assert (site_indices == tuple(lattice.site_indices) == tuple(
        range(x_dimension * y_dimension)))

    tuple(
        itertools.product(range(x_dimension), range(y_dimension), range(n_dofs),
                          range(n_spin_values)))

    spin_orbital_linear_indices = tuple(
        lattice.to_spin_orbital_index(*indices)
        for indices in itertools.product(site_indices, range(n_dofs),
                                         range(n_spin_values)))
    assert spin_orbital_linear_indices == tuple(range(lattice.n_spin_orbitals))

    for i, ii in zip(range(lattice.n_sites), lattice.site_pairs_iter('onsite')):
        assert ii == (i, i)

    n_neighbor_pairs = 2 * ((x_dimension * (y_dimension - (not periodic))) +
                            ((x_dimension - (not periodic)) * y_dimension))
    neighbor_pairs = tuple(lattice.site_pairs_iter('neighbor'))
    assert (2 * len(tuple(lattice.site_pairs_iter('neighbor', False))) ==
            len(neighbor_pairs) == n_neighbor_pairs)
    for i, j in neighbor_pairs:
        assert sum(lattice.delta_mag(i, j, True)) == 1
        assert lattice.manhattan_distance(i, j, True) == 1

    assert len(tuple(lattice.dof_pairs_iter(False))) == \
        n_dofs * (n_dofs + 1) / 2
    assert len(tuple(lattice.dof_pairs_iter(True))) == n_dofs * (n_dofs - 1) / 2
    spin_pairs_all = tuple(lattice.spin_pairs_iter())
    assert len(spin_pairs_all) == n_spin_values**2
    spin_pairs_same = tuple(lattice.spin_pairs_iter(SpinPairs.SAME))
    assert spin_pairs_same == tuple((s, s) for s in range(n_spin_values))
    spin_pairs_diff = tuple(lattice.spin_pairs_iter(SpinPairs.DIFF))
    assert spin_pairs_diff == () if spinless else ((0, 1), (1, 0))


def test_hubbard_square_lattice_defaults():
    lattice = HubbardSquareLattice(3, 3)
    assert lattice.n_dofs == 1
    assert lattice.spinless == False
    assert lattice.periodic == True


@pytest.mark.parametrize('n_dofs', range(1, 4))
def test_hubbard_square_lattice_dof_validation(n_dofs):
    lattice = HubbardSquareLattice(3, 3, n_dofs=n_dofs)
    for a in range(n_dofs):
        lattice.validate_dof(a)
    for ab in itertools.combinations(range(n_dofs), 2):
        lattice.validate_dofs(ab, 2)
    for abc in itertools.combinations(range(n_dofs), 3):
        lattice.validate_dofs(abc, 3)
    with pytest.raises(ValueError):
        lattice.validate_dof(-1)
    with pytest.raises(ValueError):
        lattice.validate_dof(n_dofs)
    with pytest.raises(ValueError):
        lattice.validate_dofs((0, 0), 1)


def test_hubbard_square_lattice_edge_types():
    lattice = HubbardSquareLattice(3, 3)
    assert sorted(lattice.edge_types) == sorted(
        ('onsite', 'neighbor', 'diagonal_neighbor', 'vertical_neighbor',
         'horizontal_neighbor'))
    lattice.validate_edge_type('onsite')
    lattice.validate_edge_type('neighbor')
    lattice.validate_edge_type('diagonal_neighbor')
    lattice.validate_edge_type('vertical_neighbor')
    lattice.validate_edge_type('horizontal_neighbor')
    with pytest.raises(ValueError):
        lattice.validate_edge_type('banana')
    with pytest.raises(ValueError):
        lattice.site_pairs_iter('banana')


@pytest.mark.parametrize('d', random.sample(range(3, 10), 3))
def test_hubbard_square_lattice_1xd(d):
    for shape, periodic in itertools.product(((1, d), (d, 1)), (True, False)):
        lattice = HubbardSquareLattice(*shape, periodic=periodic)
        assert lattice.n_sites == d
        assert (len(tuple(lattice.neighbors_iter())) ==
                2 * len(tuple(lattice.neighbors_iter(False))) == 2 *
                (d - (not periodic)))
        assert (len(tuple(lattice.diagonal_neighbors_iter())) == len(
            tuple(lattice.diagonal_neighbors_iter(False))) == 0)
        assert (len(tuple(lattice.site_pairs_iter('diagonal_neighbor'))) == len(
            tuple(lattice.diagonal_neighbors_iter(False))) == 0)


@pytest.mark.parametrize('x,y',
                         (random.sample(range(3, 10), 2) for _ in range(3)))
def test_hubbard_square_lattice_neighbors(x, y):
    for periodic in (True, False):
        lattice = HubbardSquareLattice(x, y, periodic=periodic)
        n_horizontal_neighbors = 2 * y * (x - (not periodic))
        assert (len(tuple(
            lattice.horizontal_neighbors_iter())) == n_horizontal_neighbors)
        n_vertical_neighbors = 2 * x * (y - (not periodic))
        assert (len(tuple(
            lattice.vertical_neighbors_iter())) == n_vertical_neighbors)
        n_diagonal_neighbors = 4 * (x - (not periodic)) * (y - (not periodic))
        assert (len(tuple(
            lattice.diagonal_neighbors_iter())) == n_diagonal_neighbors)


@pytest.mark.parametrize('d', random.sample(range(3, 10), 3))
def test_hubbard_square_lattice_2xd(d):
    for shape, periodic in itertools.product(((2, d), (d, 2)), (True, False)):
        lattice = HubbardSquareLattice(*shape, periodic=periodic)
        assert lattice.n_sites == 2 * d
        n_neighbor_pairs = 2 * (2 * (d - (not periodic)) + d)
        assert (len(tuple(lattice.neighbors_iter())) == 2 *
                len(tuple(lattice.neighbors_iter(False))) == n_neighbor_pairs)
        n_diagonal_neighbor_pairs = 4 * (d - (not periodic))
        assert (len(tuple(lattice.diagonal_neighbors_iter())) ==
                2 * len(tuple(lattice.diagonal_neighbors_iter(False))) ==
                n_diagonal_neighbor_pairs)


def test_hubbard_square_lattice_2x2():
    for periodic in (True, False):
        lattice = HubbardSquareLattice(2, 2, periodic=periodic)
        assert lattice.n_sites == 4
        assert len(tuple(lattice.neighbors_iter(False))) == 4
        assert len(tuple(lattice.neighbors_iter())) == 8
        assert len(tuple(lattice.diagonal_neighbors_iter(False))) == 2
        assert len(tuple(lattice.diagonal_neighbors_iter())) == 4


@pytest.mark.parametrize('kwargs', [{
    'x_dimension': random.randrange(1, 10),
    'y_dimension': random.randrange(1, 10),
    'n_dofs': random.randrange(1, 10),
    'spinless': random.choice((True, False)),
    'periodic': random.choice((True, False))
} for _ in range(5)])
def test_hubbard_square_lattice_repr(kwargs):
    lattice = HubbardSquareLattice(**kwargs)
    params = ('x_dimension', 'y_dimension', 'n_dofs', 'spinless', 'periodic')
    param_template = ', '.join('{0}={{{0}}}'.format(param) for param in params)
    param_str = param_template.format(**kwargs)
    assert repr(lattice) == 'HubbardSquareLattice({})'.format(param_str)


def test_spin_pairs_iter():
    spinful_lattice = HubbardSquareLattice(3, 3)

    with pytest.raises(ValueError):
        spinful_lattice.spin_pairs_iter(10)

    assert (tuple(spinful_lattice.spin_pairs_iter(SpinPairs.ALL,
                                                  True)) == ((0, 0), (0, 1),
                                                             (1, 0), (1, 1)))
    assert (tuple(spinful_lattice.spin_pairs_iter(SpinPairs.ALL,
                                                  False)) == ((0, 0), (0, 1),
                                                              (1, 1)))
    assert (tuple(spinful_lattice.spin_pairs_iter(SpinPairs.SAME, True)) ==
            tuple(spinful_lattice.spin_pairs_iter(SpinPairs.SAME, False)) ==
            ((0, 0), (1, 1)))
    assert (tuple(spinful_lattice.spin_pairs_iter(SpinPairs.DIFF,
                                                  True)) == ((0, 1), (1, 0)))
    assert (tuple(spinful_lattice.spin_pairs_iter(SpinPairs.DIFF,
                                                  False)) == ((0, 1),))

    spinless_lattice = HubbardSquareLattice(3, 3, spinless=True)
    assert (tuple(spinless_lattice.spin_pairs_iter(SpinPairs.ALL, True)) ==
            tuple(spinless_lattice.spin_pairs_iter(SpinPairs.ALL, False)) ==
            tuple(spinless_lattice.spin_pairs_iter(SpinPairs.SAME, True)) ==
            tuple(spinless_lattice.spin_pairs_iter(SpinPairs.SAME, False)) ==
            ((0, 0),))
    assert (tuple(spinless_lattice.spin_pairs_iter(
        SpinPairs.DIFF, True)) == tuple(
            spinless_lattice.spin_pairs_iter(SpinPairs.DIFF, False)) == tuple())



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools
import numpy
import scipy
import scipy.linalg


# Exceptions.
class OrbitalSpecificationError(Exception):
    pass


class Grid:
    """
    A multi-dimension grid of points with an assigned length scale.

    This grid acts as a helper class for parallelpiped super cells.  It
        tracks a mapping from indices to grid points and stores the associated
        reciprocal lattice with respect to the original real-space lattice.
        This enables calculations with non-trivial unit cells.

    Attributes:
        dimensions (int): Number of spatial dimensions the grid occupys
        length (tuple of ints): d-length tuple specifying number of points
            along each dimension.
        shifts (list of ints): Integer shifts in position to center grid.
        scale (ndarray): Vectors defining the super cell being simulated,
            vectors are stored as columns in the matrix.
        volume (float): Total volume of the supercell parallelpiped.
        num_points (int): Total number of points in the grid.
        reciprocal_scale (ndarray): Vectors defining the reciprocal lattice.
            The vectors are stored as the columns in the matrix.
    """

    def __init__(self, dimensions, length, scale):
        """
        Args:
            dimensions (int): The number of dimensions the grid lives in.
            length (int or tuple): The number of points along each grid axis
                that will be taken in both reciprocal and real space.
                If tuple, it is read for each dimension, otherwise assumed
                uniform.
            scale (float or ndarray): The total length of each grid dimension.
                If a float is passed, the uniform cubic unit cell is assumed.
                For an ndarray, dimensions independent vectors of the correct
                dimension must be passed.  We assume column vectors define
                the supercell vectors.
        """
        if not isinstance(dimensions, int) or dimensions <= 0:
            raise ValueError(
                'dimensions must be a positive int but was {} {}'.format(
                    type(dimensions), repr(dimensions)))
        if ((not isinstance(length, int) or length < 0) and
            (not isinstance(length, tuple)) and (not isinstance(length, list))):
            raise ValueError('length must be a non-negative int or tuple '
                             'but was {} {}'.format(type(length), repr(length)))
        if ((not isinstance(scale, float) or not scale > 0) and
            (not isinstance(scale, numpy.ndarray))):
            raise ValueError(
                'scale must be a positive float or ndarray but was '
                '{} {}'.format(type(scale), repr(scale)))

        self.dimensions = dimensions

        # If single integer, assume uniform
        if isinstance(length, int):
            self.length = (length,) * dimensions
        else:
            self.length = length

        self.shifts = [self.length[i] // 2 for i in range(dimensions)]

        # If single float, construct cubic unit cell
        if isinstance(scale, float):
            self.scale = numpy.diag([scale] * self.dimensions)
        else:
            self.scale = scale

        # Compute the volume of the super cell
        self.volume = numpy.abs(scipy.linalg.det(self.scale))

        # Compute total number of points
        self.num_points = numpy.prod(self.length)

        # Compute the reciprocal lattice basis
        self.reciprocal_scale = 2 * numpy.pi * scipy.linalg.inv(self.scale).T

    def volume_scale(self):
        """
        Returns:
            float: The volume of a length-scale hypercube within the grid.
        """
        return self.volume

    def all_points_indices(self):
        """
        Returns:
            iterable[tuple[int]]:
                The index-coordinate tuple of each point in the grid.
        """
        return itertools.product(
            *[range(self.length[i]) for i in range(self.dimensions)])

    def position_vector(self, position_indices):
        """Given grid point coordinate, return position vector with dimensions.

        Args:
            position_indices (int|iterable[int]):
                List or tuple of integers giving grid point coordinate.
                Allowed values are ints in [0, grid_length).

        Returns:
            position_vector (numpy.ndarray[float])
        """
        # Raise exceptions.
        if isinstance(position_indices, int):
            position_indices = [position_indices]
        if not all(0 <= e < self.length[i]
                   for i, e in enumerate(position_indices)):
            raise OrbitalSpecificationError(
                'Position indices must be integers in [0, grid_length).')

        # Compute position vector
        vector = sum([
            (float(n - self.shifts[i]) / self.length[i]) * self.scale[:, i]
            for i, n in enumerate(position_indices)
        ])
        return vector

    def momentum_vector(self, momentum_indices, periodic=True):
        """Given grid point coordinate, return momentum vector with dimensions.

        Args:
            momentum_indices (list): integers giving momentum
                indices. Allowed values are ints in [0, grid_length).
            periodic (bool): Wrap the momentum indices according to periodicity

            Returns:
                momentum_vector: A numpy array giving the momentum vector with
                    dimensions.
        """
        # Raise exceptions.
        if isinstance(momentum_indices, int):
            momentum_indices = [momentum_indices]
        if (not all(0 <= e < self.length[i]
                    for i, e in enumerate(momentum_indices))):
            raise OrbitalSpecificationError(
                'Momentum indices must be integers in [0, grid_length).')

        # Compute momentum vector.
        momentum_ints = self.index_to_momentum_ints(momentum_indices)
        vector = self.momentum_ints_to_value(momentum_ints, periodic)

        return vector

    def index_to_momentum_ints(self, index):
        """
        Args:
            index (tuple): d-dimensional tuple specifying index in the grid
        Returns:
            Integer momentum vector
        """
        # Set baseline for grid between [-N//2, N//2]
        momentum_int = [
            index[i] - self.shifts[i] for i in range(self.dimensions)
        ]

        return numpy.array(momentum_int, dtype=int)

    def momentum_ints_to_index(self, momentum_ints):
        """
        Args:
            momentum_ints (tuple): d-dimensional tuple momentum integers
        Returns:
            d-dimensional tuples of indices
        """

        indices = momentum_ints

        # Shift to indices
        indices = [n + self.shifts[i] for i, n in enumerate(indices)]

        # Wrap dimensions
        indices = [n % self.length[i] for i, n in enumerate(indices)]

        return indices

    def momentum_ints_to_value(self, momentum_ints, periodic=True):
        """
        Args:
            momentum_ints (tuple): d-dimensional tuple momentum integers
            periodic (bool): Alias the momentum
        Returns:
            ndarray containing the momentum vector.

        """
        # Alias the higher momentum modes
        if periodic:
            momentum_ints = self.index_to_momentum_ints(
                self.momentum_ints_to_index(momentum_ints))

        momentum_vector = sum([
            n * self.reciprocal_scale[:, i] for i, n in enumerate(momentum_ints)
        ])
        return momentum_vector

    def orbital_id(self, grid_coordinates, spin=None):
        """Return the tensor factor of a orbital
        with given coordinates and spin.

        Args:
            grid_coordinates: List or tuple of ints giving coordinates of grid
                element. Acceptable to provide an int(instead of tuple or list)
                for 1D case.
            spin (bool): 0 means spin down and 1 means spin up.
                If None, assume spinless model.

        Returns:
            tensor_factor (int):
                tensor factor associated with provided orbital label.
        """
        # Initialize.
        if isinstance(grid_coordinates, int):
            grid_coordinates = [grid_coordinates]

        # Loop through dimensions of coordinate tuple.
        tensor_factor = 0
        for dimension, grid_coordinate in enumerate(grid_coordinates):

            # Make sure coordinate is an integer in the correct bounds.
            if (isinstance(grid_coordinate, int) and
                    grid_coordinate < self.length[dimension]):
                tensor_factor += (grid_coordinate *
                                  int(numpy.product(self.length[:dimension])))
            else:
                # Raise for invalid model.
                raise OrbitalSpecificationError(
                    'Invalid orbital coordinates provided.')

        # Account for spin and return.
        if spin is None:
            return tensor_factor
        else:
            tensor_factor *= 2
            tensor_factor += spin
            return tensor_factor

    def grid_indices(self, qubit_id, spinless):
        """This function is the inverse of orbital_id.

        Args:
            qubit_id (int): The tensor factor to map to grid indices.
            spinless (bool): Whether to use the spinless model or not.

        Returns:
            grid_indices (numpy.ndarray[int]):
                The location of the qubit on the grid.
        """
        if not (numpy.product(self.length) * (2 - spinless) > qubit_id >= 0):
            raise OrbitalSpecificationError('Invalid qubit_id provided.')

        # Remove spin degree of freedom if it exists.
        orbital_id = qubit_id

        if not spinless:
            orbital_id //= 2

        # Get grid indices.
        grid_indices = []
        for dimension in range(self.dimensions):
            remainder = (orbital_id %
                         int(numpy.product(self.length[:dimension + 1])))
            grid_index = (remainder //
                          int(numpy.product(self.length[:dimension])))
            grid_indices += [grid_index]
        return grid_indices

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.dimensions == other.dimensions and
                (self.scale == other.scale).all() and
                self.length == other.length)

    def __ne__(self, other):
        return not self == other



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module to test unitary coupled cluster operators."""

import os
import unittest
import numpy
from numpy.random import randn
import scipy

from openfermion.config import DATA_DIRECTORY
from openfermion.chem import MolecularData
from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import (get_fermion_operator,
                                                  jordan_wigner, normal_ordered)

from openfermion.hamiltonians.special_operators import (s_squared_operator,
                                                        sz_operator)
from openfermion.linalg.sparse_tools import (expectation, jordan_wigner_sparse,
                                             jw_hartree_fock_state,
                                             get_sparse_operator)

from openfermion.utils import (commutator, count_qubits, hermitian_conjugated)
from openfermion.circuits.unitary_cc import (
    uccsd_generator, uccsd_singlet_generator,
    uccsd_singlet_get_packed_amplitudes, uccsd_singlet_paramsize)


class UnitaryCC(unittest.TestCase):

    def test_uccsd_anti_hermitian(self):
        """Test operators are anti-Hermitian independent of inputs"""
        test_orbitals = 4

        single_amplitudes = randn(*(test_orbitals,) * 2)
        double_amplitudes = randn(*(test_orbitals,) * 4)

        generator = uccsd_generator(single_amplitudes, double_amplitudes)
        conj_generator = hermitian_conjugated(generator)

        self.assertEqual(generator, -1. * conj_generator)

    def test_uccsd_singlet_anti_hermitian(self):
        """Test that the singlet version is anti-Hermitian"""
        test_orbitals = 8
        test_electrons = 4

        packed_amplitude_size = uccsd_singlet_paramsize(test_orbitals,
                                                        test_electrons)

        packed_amplitudes = randn(int(packed_amplitude_size))

        generator = uccsd_singlet_generator(packed_amplitudes, test_orbitals,
                                            test_electrons)

        conj_generator = hermitian_conjugated(generator)

        self.assertEqual(generator, -1. * conj_generator)

    def test_uccsd_singlet_symmetries(self):
        """Test that the singlet generator has the correct symmetries."""
        test_orbitals = 8
        test_electrons = 4

        packed_amplitude_size = uccsd_singlet_paramsize(test_orbitals,
                                                        test_electrons)
        packed_amplitudes = randn(int(packed_amplitude_size))
        generator = uccsd_singlet_generator(packed_amplitudes, test_orbitals,
                                            test_electrons)

        # Construct symmetry operators
        sz = sz_operator(test_orbitals)
        s_squared = s_squared_operator(test_orbitals)

        # Check the symmetries
        comm_sz = normal_ordered(commutator(generator, sz))
        comm_s_squared = normal_ordered(commutator(generator, s_squared))
        zero = FermionOperator()

        self.assertEqual(comm_sz, zero)
        self.assertEqual(comm_s_squared, zero)

    def test_uccsd_singlet_builds(self):
        """Test specific builds of the UCCSD singlet operator"""
        # Build 1
        n_orbitals = 4
        n_electrons = 2
        n_params = uccsd_singlet_paramsize(n_orbitals, n_electrons)
        self.assertEqual(n_params, 2)

        initial_amplitudes = [1., 2.]

        generator = uccsd_singlet_generator(initial_amplitudes, n_orbitals,
                                            n_electrons)

        test_generator = (FermionOperator("2^ 0", 1.) +
                          FermionOperator("0^ 2", -1.) +
                          FermionOperator("3^ 1", 1.) +
                          FermionOperator("1^ 3", -1.) +
                          FermionOperator("2^ 0 3^ 1", 4.) +
                          FermionOperator("1^ 3 0^ 2", -4.))

        self.assertEqual(normal_ordered(test_generator),
                         normal_ordered(generator))

        # Build 2
        n_orbitals = 6
        n_electrons = 2

        n_params = uccsd_singlet_paramsize(n_orbitals, n_electrons)
        self.assertEqual(n_params, 5)

        initial_amplitudes = numpy.arange(1, n_params + 1, dtype=float)
        generator = uccsd_singlet_generator(initial_amplitudes, n_orbitals,
                                            n_electrons)

        test_generator = (
            FermionOperator("2^ 0", 1.) + FermionOperator("0^ 2", -1) +
            FermionOperator("3^ 1", 1.) + FermionOperator("1^ 3", -1.) +
            FermionOperator("4^ 0", 2.) + FermionOperator("0^ 4", -2) +
            FermionOperator("5^ 1", 2.) + FermionOperator("1^ 5", -2.) +
            FermionOperator("2^ 0 3^ 1", 6.) +
            FermionOperator("1^ 3 0^ 2", -6.) +
            FermionOperator("4^ 0 5^ 1", 8.) +
            FermionOperator("1^ 5 0^ 4", -8.) +
            FermionOperator("2^ 0 5^ 1", 5.) +
            FermionOperator("1^ 5 0^ 2", -5.) +
            FermionOperator("4^ 0 3^ 1", 5.) +
            FermionOperator("1^ 3 0^ 4", -5.) +
            FermionOperator("2^ 0 4^ 0", 5.) +
            FermionOperator("0^ 4 0^ 2", -5.) +
            FermionOperator("3^ 1 5^ 1", 5.) +
            FermionOperator("1^ 5 1^ 3", -5.))

        self.assertEqual(normal_ordered(test_generator),
                         normal_ordered(generator))

    def test_sparse_uccsd_generator_numpy_inputs(self):
        """Test numpy ndarray inputs to uccsd_generator that are sparse"""
        test_orbitals = 30
        sparse_single_amplitudes = numpy.zeros((test_orbitals, test_orbitals))
        sparse_double_amplitudes = numpy.zeros(
            (test_orbitals, test_orbitals, test_orbitals, test_orbitals))

        sparse_single_amplitudes[3, 5] = 0.12345
        sparse_single_amplitudes[12, 4] = 0.44313

        sparse_double_amplitudes[0, 12, 6, 2] = 0.3434
        sparse_double_amplitudes[1, 4, 6, 13] = -0.23423

        generator = uccsd_generator(sparse_single_amplitudes,
                                    sparse_double_amplitudes)

        test_generator = (0.12345 * FermionOperator("3^ 5") +
                          (-0.12345) * FermionOperator("5^ 3") +
                          0.44313 * FermionOperator("12^ 4") +
                          (-0.44313) * FermionOperator("4^ 12") +
                          0.3434 * FermionOperator("0^ 12 6^ 2") +
                          (-0.3434) * FermionOperator("2^ 6 12^ 0") +
                          (-0.23423) * FermionOperator("1^ 4 6^ 13") +
                          0.23423 * FermionOperator("13^ 6 4^ 1"))
        self.assertEqual(test_generator, generator)

    def test_sparse_uccsd_generator_list_inputs(self):
        """Test list inputs to uccsd_generator that are sparse"""
        sparse_single_amplitudes = [[[3, 5], 0.12345], [[12, 4], 0.44313]]
        sparse_double_amplitudes = [[[0, 12, 6, 2], 0.3434],
                                    [[1, 4, 6, 13], -0.23423]]

        generator = uccsd_generator(sparse_single_amplitudes,
                                    sparse_double_amplitudes)

        test_generator = (0.12345 * FermionOperator("3^ 5") +
                          (-0.12345) * FermionOperator("5^ 3") +
                          0.44313 * FermionOperator("12^ 4") +
                          (-0.44313) * FermionOperator("4^ 12") +
                          0.3434 * FermionOperator("0^ 12 6^ 2") +
                          (-0.3434) * FermionOperator("2^ 6 12^ 0") +
                          (-0.23423) * FermionOperator("1^ 4 6^ 13") +
                          0.23423 * FermionOperator("13^ 6 4^ 1"))
        self.assertEqual(test_generator, generator)

    def test_uccsd_singlet_get_packed_amplitudes(self):
        test_orbitals = 6
        test_electrons = 2
        sparse_single_amplitudes = numpy.zeros((test_orbitals, test_orbitals))
        sparse_double_amplitudes = numpy.zeros(
            (test_orbitals, test_orbitals, test_orbitals, test_orbitals))

        sparse_single_amplitudes[2, 0] = 0.12345
        sparse_single_amplitudes[3, 1] = 0.12345

        sparse_double_amplitudes[2, 0, 3, 1] = 0.9

        sparse_double_amplitudes[2, 0, 4, 0] = 0.3434
        sparse_double_amplitudes[5, 1, 3, 1] = 0.3434

        packed_amplitudes = uccsd_singlet_get_packed_amplitudes(
            sparse_single_amplitudes, sparse_double_amplitudes, test_orbitals,
            test_electrons)

        self.assertEqual(len(packed_amplitudes), 5)

        self.assertTrue(
            numpy.allclose(packed_amplitudes,
                           numpy.array([0.12345, 0., 0.9, 0., 0.3434])))

    def test_ucc_h2(self):
        geometry = [('H', (0., 0., 0.)), ('H', (0., 0., 0.7414))]
        basis = 'sto-3g'
        multiplicity = 1
        filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        self.molecule = MolecularData(geometry,
                                      basis,
                                      multiplicity,
                                      filename=filename)
        self.molecule.load()

        # Get molecular Hamiltonian.
        self.molecular_hamiltonian = self.molecule.get_molecular_hamiltonian()

        # Get FCI RDM.
        self.fci_rdm = self.molecule.get_molecular_rdm(use_fci=1)

        # Get explicit coefficients.
        self.nuclear_repulsion = self.molecular_hamiltonian.constant
        self.one_body = self.molecular_hamiltonian.one_body_tensor
        self.two_body = self.molecular_hamiltonian.two_body_tensor

        # Get fermion Hamiltonian.
        self.fermion_hamiltonian = normal_ordered(
            get_fermion_operator(self.molecular_hamiltonian))

        # Get qubit Hamiltonian.
        self.qubit_hamiltonian = jordan_wigner(self.fermion_hamiltonian)

        # Get the sparse matrix.
        self.hamiltonian_matrix = get_sparse_operator(
            self.molecular_hamiltonian)
        # Test UCCSD for accuracy against FCI using loaded t amplitudes.
        ucc_operator = uccsd_generator(self.molecule.ccsd_single_amps,
                                       self.molecule.ccsd_double_amps)

        hf_state = jw_hartree_fock_state(self.molecule.n_electrons,
                                         count_qubits(self.qubit_hamiltonian))
        uccsd_sparse = jordan_wigner_sparse(ucc_operator)
        uccsd_state = scipy.sparse.linalg.expm_multiply(uccsd_sparse, hf_state)
        expected_uccsd_energy = expectation(self.hamiltonian_matrix,
                                            uccsd_state)
        self.assertAlmostEqual(expected_uccsd_energy,
                               self.molecule.fci_energy,
                               places=4)
        print("UCCSD ENERGY: {}".format(expected_uccsd_energy))

        # Test CCSD for precise match against FCI using loaded t amplitudes.
        ccsd_operator = uccsd_generator(self.molecule.ccsd_single_amps,
                                        self.molecule.ccsd_double_amps,
                                        anti_hermitian=False)

        ccsd_sparse_r = jordan_wigner_sparse(ccsd_operator)
        ccsd_sparse_l = jordan_wigner_sparse(
            -hermitian_conjugated(ccsd_operator))

        ccsd_state_r = scipy.sparse.linalg.expm_multiply(
            ccsd_sparse_r, hf_state)
        ccsd_state_l = scipy.sparse.linalg.expm_multiply(
            ccsd_sparse_l, hf_state)
        expected_ccsd_energy = ccsd_state_l.conjugate().dot(
            self.hamiltonian_matrix.dot(ccsd_state_r))
        self.assertAlmostEqual(expected_ccsd_energy, self.molecule.fci_energy)

    def test_ucc_h2_singlet(self):
        geometry = [('H', (0., 0., 0.)), ('H', (0., 0., 0.7414))]
        basis = 'sto-3g'
        multiplicity = 1
        filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        self.molecule = MolecularData(geometry,
                                      basis,
                                      multiplicity,
                                      filename=filename)
        self.molecule.load()

        # Get molecular Hamiltonian.
        self.molecular_hamiltonian = self.molecule.get_molecular_hamiltonian()

        # Get FCI RDM.
        self.fci_rdm = self.molecule.get_molecular_rdm(use_fci=1)

        # Get explicit coefficients.
        self.nuclear_repulsion = self.molecular_hamiltonian.constant
        self.one_body = self.molecular_hamiltonian.one_body_tensor
        self.two_body = self.molecular_hamiltonian.two_body_tensor

        # Get fermion Hamiltonian.
        self.fermion_hamiltonian = normal_ordered(
            get_fermion_operator(self.molecular_hamiltonian))

        # Get qubit Hamiltonian.
        self.qubit_hamiltonian = jordan_wigner(self.fermion_hamiltonian)

        # Get the sparse matrix.
        self.hamiltonian_matrix = get_sparse_operator(
            self.molecular_hamiltonian)
        # Test UCCSD for accuracy against FCI using loaded t amplitudes.
        ucc_operator = uccsd_generator(self.molecule.ccsd_single_amps,
                                       self.molecule.ccsd_double_amps)

        hf_state = jw_hartree_fock_state(self.molecule.n_electrons,
                                         count_qubits(self.qubit_hamiltonian))
        uccsd_sparse = jordan_wigner_sparse(ucc_operator)
        uccsd_state = scipy.sparse.linalg.expm_multiply(uccsd_sparse, hf_state)
        expected_uccsd_energy = expectation(self.hamiltonian_matrix,
                                            uccsd_state)
        self.assertAlmostEqual(expected_uccsd_energy,
                               self.molecule.fci_energy,
                               places=4)
        print("UCCSD ENERGY: {}".format(expected_uccsd_energy))

        # Test CCSD singlet for precise match against FCI using loaded t
        # amplitudes
        packed_amplitudes = uccsd_singlet_get_packed_amplitudes(
            self.molecule.ccsd_single_amps, self.molecule.ccsd_double_amps,
            self.molecule.n_qubits, self.molecule.n_electrons)
        ccsd_operator = uccsd_singlet_generator(packed_amplitudes,
                                                self.molecule.n_qubits,
                                                self.molecule.n_electrons,
                                                anti_hermitian=False)

        ccsd_sparse_r = jordan_wigner_sparse(ccsd_operator)
        ccsd_sparse_l = jordan_wigner_sparse(
            -hermitian_conjugated(ccsd_operator))

        ccsd_state_r = scipy.sparse.linalg.expm_multiply(
            ccsd_sparse_r, hf_state)
        ccsd_state_l = scipy.sparse.linalg.expm_multiply(
            ccsd_sparse_l, hf_state)
        expected_ccsd_energy = ccsd_state_l.conjugate().dot(
            self.hamiltonian_matrix.dot(ccsd_state_r))
        self.assertAlmostEqual(expected_ccsd_energy, self.molecule.fci_energy)

    def test_value_error_for_odd_n_qubits(self):
        # Pass odd n_qubits to singlet generators
        with self.assertRaises(ValueError):
            _ = uccsd_singlet_paramsize(3, 4)

    def test_value_error_bad_amplitudes(self):
        with self.assertRaises(ValueError):
            _ = uccsd_singlet_generator([1.], 3, 4)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import abc
import itertools
from typing import (cast, Dict, Optional, Sequence, Tuple, TYPE_CHECKING, Union)

import numpy as np
import scipy.linalg as la
import sympy
import cirq

import openfermion.ops as ops
from openfermion.transforms import get_interaction_operator
from openfermion.linalg import jordan_wigner_sparse
from openfermion.utils import hermitian_conjugated

if TYPE_CHECKING:
    import openfermion


def _arg(x):
    if x == 0:
        return 0
    if cirq.is_parameterized(x):
        return sympy.arg(x)
    return np.angle(x)


def _canonicalize_weight(w):
    if w == 0:
        return (0, 0)
    if cirq.is_parameterized(w):
        return (cirq.PeriodicValue(abs(w), 2 * sympy.pi), sympy.arg(w))
    period = 2 * np.pi
    return (np.round((w.real % period) if (w == np.real(w)) else
                     (abs(w) % period) * w / abs(w), 8), 0)


def state_swap_eigen_component(x: str, y: str, sign: int = 1, angle: float = 0):
    """The +/- eigen-component of the operation that swaps states x and y.

    For example, state_swap_eigen_component('01', '10', ±1) with angle θ returns
        ┌                               ┐
        │0, 0,           0,            0│
        │0, 0.5,         ±0.5 e^{-iθ}, 0│
        │0, ±0.5 e^{iθ}, 0.5,          0│
        │0, 0,           0,            0│
        └                               ┘

    Args:
        x: The first state to swap, as a bitstring.
        y: The second state to swap, as a bitstring. Must have high index than
            x.
        sign: The sign of the off-diagonal elements (indicated by +/-1).
        angle: The phase of the complex off-diagonal elements. Defaults to 0.

    Returns: The eigen-component.

    Raises:
        ValueError:
            * x and y have different lengths
            * x or y contains a character other than '0' and '1'
            * x and y are the same
            * sign is not -1 or 1
        TypeError: x or y is not a string
    """
    if not (isinstance(x, str) and isinstance(y, str)):
        raise TypeError('not (isinstance(x, str) and isinstance(y, str))')
    if len(x) != len(y):
        raise ValueError('len(x) != len(y)')
    if set(x).union(y).difference('01'):
        raise ValueError('Arguments must be 0-1 strings.')
    if x == y:
        raise ValueError('x == y')
    if sign not in (-1, 1):
        raise ValueError('sign not in (-1, 1)')

    dim = 2**len(x)
    i, j = int(x, 2), int(y, 2)

    component = np.zeros((dim, dim), dtype=np.complex128)
    component[i, i] = component[j, j] = 0.5
    component[j, i] = sign * 0.5 * 1j**(angle * 2 / np.pi)
    component[i, j] = sign * 0.5 * 1j**(-angle * 2 / np.pi)
    return component


def fermionic_simulation_gates_from_interaction_operator(
        operator: 'openfermion.InteractionOperator'):
    r"""
    Given $H = \sum_{I \subset [n]} H_I$, returns gates
    $\left\{G_I\right\} = \left\{e^{i H_I\right\}$.

    Each term $H_I$ is the sum of all terms in $H$ that involve exactly the
    orbitals $I$.

    Args:
        operator: The interaction operator ($H$).

    Returns: A dict from tuples of mode indices to gates.
    """
    n_qubits = operator.n_qubits

    gates: Dict[Tuple[int, ...], cirq.Gate] = {}

    if operator.constant:
        gates[()] = operator.constant
    for p in range(n_qubits):
        coeff = operator.one_body_tensor[p, p]
        if coeff:
            gates[(p,)] = cirq.Z**(coeff / np.pi)
    for modes in itertools.combinations(range(n_qubits), 2):
        gate: Optional[InteractionOperatorFermionicGate] = (
            QuadraticFermionicSimulationGate.from_interaction_operator(
                operator=operator, modes=modes))
        if gate:
            gates[modes] = gate
    for modes in itertools.combinations(range(n_qubits), 3):
        gate = CubicFermionicSimulationGate.from_interaction_operator(
            operator=operator, modes=modes)
        if gate:
            gates[modes] = gate
    for modes in itertools.combinations(range(n_qubits), 4):
        gate = QuarticFermionicSimulationGate.from_interaction_operator(
            operator=operator, modes=modes)
        if gate:
            gates[modes] = gate
    return gates


def sum_of_interaction_operator_gate_generators(
        n_modes: int,
        gates: Dict[Tuple[int, ...], Union[float, cirq.Gate]],
) -> 'openfermion.InteractionOperator':
    """The interaction operator that is the sum of the generators of the
    specified fermionic simulation gates.

    The gates are specified as a dictionary whose items are (indices, gate),
    where
        * indices is a tuple of ints specifying the modes on which the gate
          acts;
        * gate is one of type
          - float, which is interpreted as a constant, regardless of the
            indices,
          - cirq.ZPowGate, which is interpreted as a "linear" fermionic
            simulation gate,
          - openfermion.InteractionOperatorFermionicGate.

    Args:
        n_modes: The number of modes.
        gates: The gates.

    Returns:
        The interaction operator.
    """
    # assumes gate indices in JW order
    operator = ops.InteractionOperator.zero(n_modes)

    for indices, gate in gates.items():
        if not indices:
            operator.constant += gate
        elif isinstance(gate, cirq.ZPowGate):
            coeff = gate._exponent * np.pi
            operator.constant += gate._exponent * gate._global_shift * np.pi
            operator.one_body_tensor[indices * 2] += coeff
        elif isinstance(gate, InteractionOperatorFermionicGate):
            gate.interaction_operator_generator(operator=operator,
                                                modes=indices)
        else:
            raise TypeError(f'Gate type {gate} not supported.')

    return operator


@cirq.value_equality(approximate=True)
class ParityPreservingFermionicGate(cirq.Gate, metaclass=abc.ABCMeta):
    r"""The Jordan-Wigner transform of :math:`\exp(-i H)` for a fermionic
    Hamiltonian :math:`H`.

    Each subclass corresponds to a set of generators :math:`\{G_i\}`
    corresponding to the family of Hamiltonians :math:`\sum_i w_i G_i +
    \text{h.c.}`, where the weights :math:`w_i \in \mathbb C` are specified by
    the instance.

    The Jordan-Wigner mapping maps the fermionic modes :math:`(0, \ldots, n -
    1)` to qubits :math:`(0, \ldots, n - 1)`, respectively.

    Each generator :math:`G_i` must be a linear combination of fermionic
    monomials consisting of an even number of creation/annihilation operators.
    This is so that the Jordan-Wigner transform acts only on the gate's qubits,
    even when the fermionic modes are offset as part of a larger Jordan-Wigner
    string.
    """

    def __init__(
            self,
            weights: Optional[Tuple[complex, ...]] = None,
            absorb_exponent: bool = False,
            exponent: cirq.TParamVal = 1.0,
            global_shift: float = 0.0,
    ) -> None:
        """A fermionic interaction.

        Args:
            weights: The weights of the terms in the Hamiltonian.
            absorb_exponent: Whether to absorb the given exponent into the
                weights. If true, the exponent of the return gate is `1`.
                Defaults to `False`.
        """
        if weights is None:
            weights = (1.,) * self.num_weights()
        self.weights = weights

        self._exponent = exponent
        self._global_shift = global_shift
        self._canonical_exponent_cached = None

        if absorb_exponent:
            self.absorb_exponent_into_weights()

    @staticmethod
    @abc.abstractmethod
    def fermion_generator_components() -> Tuple['openfermion.FermionOperator']:
        r"""The FermionOperators :math:`(G_i)_i` such that the gate's fermionic
        generator is :math:`\sum_i w_i G_i + \text{h.c.}` where :math:`(w_i)_i`
        are the gate's weights."""

    @abc.abstractmethod
    def fswap(self, i: int):
        """Update the weights of the gate to effect conjugation by an FSWAP on
        the i-th and (i+1)th qubits."""

    @classmethod
    def num_weights(cls) -> int:
        """The number of parameters (weights) in the generator."""
        return len(cls.fermion_generator_components())

    @property
    def qubit_generator_matrix(self) -> np.ndarray:
        """The matrix G such that the gate's unitary is exp(-i t G) with
        exponent t."""
        return jordan_wigner_sparse(self.fermion_generator,
                                    self.num_qubits()).toarray()

    @property
    def fermion_generator(self) -> 'openfermion.FermionOperator':
        """The FermionOperator G such that the gate's unitary is exp(-i t G)
        with exponent t using the Jordan-Wigner transformation."""
        half_generator = sum((
            w * G
            for w, G in zip(self.weights, self.fermion_generator_components())),
                             ops.FermionOperator())
        return half_generator + hermitian_conjugated(half_generator)

    def _diagram_exponent(self,
                          args: cirq.CircuitDiagramInfoArgs,
                          *,
                          ignore_global_phase: bool = True):
        if not isinstance(self._exponent, (int, float)):
            return self._exponent
        result = float(self._exponent)
        if args.precision is not None:
            result = np.around(result, args.precision)
        return result

    @classmethod
    def wire_symbol(cls, use_unicode: bool):
        """The symbol to use in circuit diagrams."""
        return cls.__name__

    def _resolve_parameters_(self, resolver, recursive: bool = True):
        resolved_weights = cirq.resolve_parameters(self.weights,
                                                   resolver,
                                                   recursive=recursive)
        resolved_exponent = cirq.resolve_parameters(self._exponent,
                                                    resolver,
                                                    recursive=recursive)
        resolved_global_shift = cirq.resolve_parameters(self._global_shift,
                                                        resolver,
                                                        recursive=recursive)
        return type(self)(resolved_weights,
                          exponent=resolved_exponent,
                          global_shift=resolved_global_shift)

    def _value_equality_values_(self):
        return tuple(
            _canonicalize_weight(w * self.exponent)
            for w in list(self.weights) + [self._global_shift])

    def _is_parameterized_(self) -> bool:
        return any(
            cirq.is_parameterized(v)
            for V in self._value_equality_values_()
            for v in V)

    def absorb_exponent_into_weights(self):
        period = (2 * sympy.pi) if self._is_parameterized_() else 2 * (np.pi)
        new_weights = []
        for weight in self.weights:
            if not weight:
                new_weights.append(weight)
                continue
            old_abs = abs(weight)
            new_abs = (old_abs * self._exponent) % period
            new_weights.append(weight * new_abs / old_abs)
        self.weights = tuple(new_weights)
        self._global_shift *= self._exponent
        self._exponent = 1

    def permute(self, init_pos: Sequence[int]):
        """An in-place version of permuted."""
        I = range(self.num_qubits())
        if sorted(init_pos) != list(I):
            raise ValueError(f'{init_pos} is not a permutation of {I}.')
        curr_pos = list(init_pos)
        for i in I:
            for j in I[i % 2:-1:2]:
                if curr_pos[j] > curr_pos[j + 1]:
                    self.fswap(j)
                    curr_pos[j:j + 2] = reversed(curr_pos[j:j + 2])
        assert curr_pos == list(I)

    def permuted(self, init_pos: Sequence[int]):
        """Returns a gate with the Jordan-Wigner ordering changed.

        If the Jordan-Wigner ordering of the original gate is given by
        init_pos, then the returned gate has Jordan-Wigner ordering
        (0, ..., n - 1), where n is the number of qubits on which the gate acts.

        Args:
            init_pos: A permutation of (0, ..., n - 1).
        """
        gate = self.__copy__()
        gate.permute(init_pos)
        return gate

    def __copy__(self):
        return type(self)(self.weights,
                          exponent=self.exponent,
                          global_shift=self._global_shift)

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs
                              ) -> cirq.CircuitDiagramInfo:
        wire_symbols = [self.wire_symbol(args.use_unicode_characters)
                       ] * self.num_qubits()
        wire_symbols[0] += f'{tuple(self.weights)}'
        exponent = self._diagram_exponent(args)
        return cirq.CircuitDiagramInfo(wire_symbols=wire_symbols,
                                       exponent=exponent)


class InteractionOperatorFermionicGate(ParityPreservingFermionicGate):
    r"""The Jordan-Wigner transform of :math:`\exp(-i H)` for a fermionic
    Hamiltonian :math:`H`, where :math:`H` is an interaction operator.

    See openfermion.ParityPreservingFermionicGate and
    openfermion.InteractionOperator for more details.
    """

    @classmethod
    @abc.abstractmethod
    def from_interaction_operator(
            cls,
            *,
            operator: 'openfermion.InteractionOperator',
            modes: Optional[Sequence[int]] = None,
    ) -> Optional['ParityPreservingFermionicGate']:
        """Constructs the gate corresponding to the specified term in the
        Hamiltonian."""

    def interaction_operator_generator(
            self,
            *,
            operator: Optional['openfermion.InteractionOperator'] = None,
            modes: Optional[Sequence[int]] = None
    ) -> 'openfermion.InteractionOperator':
        """Constructs the Hamiltonian corresponding to the gate's generator."""
        if modes is None:
            modes = tuple(range(self.num_qubits()))
        if operator is None:
            n_modes = max(modes) + 1
            operator = ops.InteractionOperator.zero(n_modes)
        else:
            n_modes = operator.n_qubits
        fermion_operator = self.fermion_generator
        return get_interaction_operator(fermion_operator, n_qubits=n_modes)


class QuadraticFermionicSimulationGate(InteractionOperatorFermionicGate,
                                       cirq.InterchangeableQubitsGate,
                                       cirq.TwoQubitGate, cirq.EigenGate):
    r"""``(w0 |10⟩⟨01| + h.c.) + w1 |11⟩⟨11|`` interaction.

    With weights :math:`(w_0, w_1)` and exponent :math:`t`, this gate's matrix
    is defined as

    .. math::
        e^{-i t H},

    where

    .. math::
        H = \left(w_0 \left| 10 \right\rangle\left\langle 01 \right| +
                \text{h.c.}\right) -
            w_1 \left| 11 \right\rangle \left\langle 11 \right|.

    This corresponds to the Jordan-Wigner transform of

    .. math::
        H = (w_0 a^{\dagger}_i a_{i+1} + \text{h.c.}) +
             w_1 a_{i}^{\dagger} a_{i+1}^{\dagger} a_{i} a_{i+1},

    where :math:`a_i` and  :math:`a_{i+1}` are the annihilation operators for
    the fermionic modes :math:`i` and :math:`(i+1)`, respectively mapped to the
    first and second qubits on which this gate acts.

    Args:
        weights: The weights of the terms in the Hamiltonian.
    """

    @classmethod
    def num_weights(cls):
        return 2

    def _decompose_(self, qubits):
        r = 2 * abs(self.weights[0]) / np.pi
        theta = _arg(self.weights[0]) / np.pi
        yield cirq.Z(qubits[0])**-theta
        yield cirq.ISwapPowGate(exponent=-r * self.exponent)(*qubits)
        yield cirq.Z(qubits[0])**theta
        yield cirq.CZPowGate(exponent=-self.weights[1] * self.exponent /
                             np.pi)(*qubits)

    def _eigen_components(self):
        components = [(0, np.diag([1, 0, 0, 0])),
                      (-self.weights[1] / np.pi, np.diag([0, 0, 0, 1]))]
        r = abs(self.weights[0]) / np.pi
        theta = 2 * _arg(self.weights[0]) / np.pi
        for s in (-1, 1):
            components.append(
                (-s * r,
                 np.array([[0, 0, 0, 0], [0, 1, s * 1j**(-theta), 0],
                           [0, s * 1j**(theta), 1, 0], [0, 0, 0, 0]]) / 2))
        return components

    def __repr__(self):
        exponent_str = ('' if self.exponent == 1 else ', exponent=' +
                        cirq._compat.proper_repr(self.exponent))
        return ('openfermion.QuadraticFermionicSimulationGate(({}){})'.format(
            ', '.join(cirq._compat.proper_repr(v) for v in self.weights),
            exponent_str))

    @property
    def qubit_generator_matrix(self):
        generator = np.zeros((4, 4), dtype=np.complex128)
        # w0 |10><01| + h.c.
        generator[2, 1] = self.weights[0]
        generator[1, 2] = self.weights[0].conjugate()
        # w1 |11><11|
        generator[3, 3] = self.weights[1]
        return generator

    @staticmethod
    def fermion_generator_components():
        return (
            ops.FermionOperator(((0, 1), (1, 0))),
            ops.FermionOperator(((0, 1), (0, 0), (1, 1), (1, 0)), 0.5),
        )

    @classmethod
    def wire_symbol(cls, use_unicode: bool):
        return '↓↑' if use_unicode else 'a*a'

    @classmethod
    def from_interaction_operator(
            cls,
            *,
            operator: 'openfermion.InteractionOperator',
            modes: Optional[Sequence[int]] = None,
    ) -> Optional['QuadraticFermionicSimulationGate']:
        if modes is None:
            modes = (0, 1)

        p, q = modes
        tunneling_coeff = operator.one_body_tensor[p, q]
        interaction_coeff = (-operator.two_body_tensor[p, q, p, q] +
                             operator.two_body_tensor[q, p, p, q] +
                             operator.two_body_tensor[p, q, q, p] -
                             operator.two_body_tensor[q, p, q, p])
        weights: Tuple[complex, complex] = (-tunneling_coeff,
                                            -interaction_coeff)
        if any(weights):
            return cls(weights)
        return None

    def interaction_operator_generator(
            self,
            *,
            operator: Optional['openfermion.InteractionOperator'] = None,
            modes: Optional[Sequence[int]] = None
    ) -> 'openfermion.InteractionOperator':
        if modes is None:
            modes = (0, 1)
        if operator is None:
            n_modes = max(modes) + 1
            operator = ops.InteractionOperator.zero(n_modes)

        weights = tuple(w * self._exponent for w in self.weights)
        operator.constant += self._exponent * self._global_shift
        operator.one_body_tensor[modes] -= weights[0]
        operator.one_body_tensor[modes[::-1]] -= weights[0].conjugate()
        operator.two_body_tensor[tuple(modes) * 2] += weights[1]
        return operator

    def fswap(self, i: int = 0):
        if i != 0:
            raise ValueError(f'{i} != 0')
        self.weights = (self.weights[0].conjugate(), self.weights[1])


class CubicFermionicSimulationGate(InteractionOperatorFermionicGate,
                                   cirq.ThreeQubitGate, cirq.EigenGate):
    r"""``w0|110⟩⟨101| + w1|110⟩⟨011| + w2|101⟩⟨011|`` + h.c. interaction.

    With weights :math:`(w_0, w_1, w_2)` and exponent :math:`t`, this gate's
    matrix is defined as

    .. math::
        e^{-i t H},

    where

    .. math::
        H = \left(w_0 \left| 110 \right\rangle\left\langle 101 \right| +
                \text{h.c.}\right) +
            \left(w_1 \left| 110 \right\rangle\left\langle 011 \right| +
                \text{h.c.}\right) +
            \left(w_2 \left| 101 \right\rangle\left\langle 011 \right| +
                \text{h.c.}\right)

    This corresponds to the Jordan-Wigner transform of

    .. math::
        H = -\left(w_0 a^{\dagger}_i a^{\dagger}_{i+1} a_{i} a_{i+2} +
                   \text{h.c.}\right) -
            \left(w_1 a^{\dagger}_i a^{\dagger}_{i+1} a_{i+1} a_{i+2} +
                  \text{h.c.}\right) -
            \left(w_2 a^{\dagger}_i a^{\dagger}_{i+2} a_{i+1} a_{i+2} +
                  \text{h.c.}\right),

    where :math:`a_i`, :math:`a_{i+1}`, :math:`a_{i+2}` are the annihilation
    operators for the fermionic modes :math:`i`, :math:`(i+1)` :math:`(i+2)`,
    respectively mapped to the three qubits on which this gate acts.

    Args:
        weights: The weights of the terms in the Hamiltonian.
    """

    @classmethod
    def num_weights(cls):
        return 3

    @classmethod
    def wire_symbol(cls, use_unicode: bool):
        return '↕↓↑' if use_unicode else 'na*a'

    def _eigen_components(self):
        components = [(0, np.diag([1, 1, 1, 0, 1, 0, 0, 1]))]
        nontrivial_part = np.zeros((3, 3), dtype=np.complex128)
        for ij, w in zip([(1, 2), (0, 2), (0, 1)], self.weights):
            nontrivial_part[ij] = w
            nontrivial_part[ij[::-1]] = w.conjugate()
        assert np.allclose(nontrivial_part, nontrivial_part.conj().T)
        eig_vals, eig_vecs = np.linalg.eigh(nontrivial_part)
        for eig_val, eig_vec in zip(eig_vals, eig_vecs.T):
            exp_factor = -eig_val / np.pi
            proj = np.zeros((8, 8), dtype=np.complex128)
            nontrivial_indices = np.array([3, 5, 6], dtype=np.intp)
            proj[nontrivial_indices[:, np.newaxis], nontrivial_indices] = (
                np.outer(eig_vec.conjugate(), eig_vec))
            components.append((exp_factor, proj))
        return components

    def __repr__(self):
        return ('openfermion.CubicFermionicSimulationGate(' + '({})'.format(
            ' ,'.join(cirq._compat.proper_repr(w) for w in self.weights)) +
                ('' if self.exponent == 1 else
                 (', exponent=' + cirq._compat.proper_repr(self.exponent))) +
                ('' if self._global_shift == 0 else
                 (', global_shift=' +
                  cirq._compat.proper_repr(self._global_shift))) + ')')

    @property
    def qubit_generator_matrix(self):
        generator = np.zeros((8, 8), dtype=np.complex128)
        # w0 |110><101| + h.c.
        generator[6, 5] = self.weights[0]
        generator[5, 6] = self.weights[0].conjugate()
        # w1 |110><011| + h.c.
        generator[6, 3] = self.weights[1]
        generator[3, 6] = self.weights[1].conjugate()
        # w2 |101><011| + h.c.
        generator[5, 3] = self.weights[2]
        generator[3, 5] = self.weights[2].conjugate()
        return generator

    @staticmethod
    def fermion_generator_components():
        return (
            ops.FermionOperator(((0, 1), (0, 0), (1, 1), (2, 0))),
            ops.FermionOperator(((0, 1), (1, 1), (1, 0), (2, 0)), -1),
            ops.FermionOperator(((0, 1), (1, 0), (2, 1), (2, 0))),
        )

    @classmethod
    def from_interaction_operator(
            cls,
            *,
            operator: 'openfermion.InteractionOperator',
            modes: Optional[Sequence[int]] = None,
    ) -> Optional['CubicFermionicSimulationGate']:
        if modes is None:
            modes = (0, 1, 2)
        i, j, k = modes
        weights = tuple(
            sgn * (operator.two_body_tensor[p, q, p, r] -
                   operator.two_body_tensor[p, q, r, p] -
                   operator.two_body_tensor[q, p, p, r] +
                   operator.two_body_tensor[q, p, r, p])
            for sgn, (p, q,
                      r) in zip([1, -1, 1], [(i, j, k), (j, i, k), (k, i, j)]))
        if any(weights):
            return cls(cast(Tuple[complex, complex, complex], weights))
        return None

    def interaction_operator_generator(
            self,
            *,
            operator: Optional['openfermion.InteractionOperator'] = None,
            modes: Optional[Sequence[int]] = None
    ) -> 'openfermion.InteractionOperator':
        if modes is None:
            modes = (0, 1, 2)
        if operator is None:
            n_modes = max(modes) + 1
            operator = ops.InteractionOperator.zero(n_modes)

        weights = tuple(w * self._exponent for w in self.weights)
        operator.constant += self._exponent * self._global_shift
        p, q, r = modes
        operator.two_body_tensor[p, q, p, r] += weights[0]
        operator.two_body_tensor[p, r, p, q] += weights[0].conjugate()
        operator.two_body_tensor[p, q, q, r] += weights[1]
        operator.two_body_tensor[q, r, p, q] += weights[1].conjugate()
        operator.two_body_tensor[p, r, q, r] += weights[2]
        operator.two_body_tensor[q, r, p, r] += weights[2].conjugate()
        return operator

    def fswap(self, i: int):
        if i == 0:
            self.weights = (-self.weights[1], -self.weights[0],
                            self.weights[2].conjugate())
        elif i == 1:
            self.weights = (self.weights[0].conjugate(), -self.weights[2],
                            -self.weights[1])
        else:
            raise ValueError(f'{i} not in (0, 1)')


class QuarticFermionicSimulationGate(InteractionOperatorFermionicGate,
                                     cirq.EigenGate):
    r"""Rotates Hamming-weight 2 states into their bitwise complements.

    With weights :math:`(w_0, w_1, w_2)` and exponent :math:`t`, this gate's
    matrix is defined as

    .. math::
        e^{-i t H},

    where

    .. math::
        H = \left(w_0 \left| 1001 \right\rangle\left\langle 0110 \right| +
                \text{h.c.}\right) +
            \left(w_1 \left| 1010 \right\rangle\left\langle 0101 \right| +
                \text{h.c.}\right) +
            \left(w_2 \left| 1100 \right\rangle\left\langle 0011 \right| +
                \text{h.c.}\right)

    This corresponds to the Jordan-Wigner transform of

    .. math::
        H = -\left(w_0 a^{\dagger}_i a^{\dagger}_{i+3} a_{i+1} a_{i+2} +
                   \text{h.c.}\right) -
            \left(w_1 a^{\dagger}_i a^{\dagger}_{i+2} a_{i+1} a_{i+3} +
                  \text{h.c.}\right) -
            \left(w_2 a^{\dagger}_i a^{\dagger}_{i+1} a_{i+2} a_{i+3} +
                  \text{h.c.}\right),

    where :math:`a_i`, ..., :math:`a_{i+3}` are the annihilation operators for
    the fermionic modes :math:`i`, ..., :math:`(i+3)`, respectively
    mapped to the four qubits on which this gate acts.


    Args:
        weights: The weights of the terms in the Hamiltonian.
    """

    @classmethod
    def num_weights(cls):
        return 3

    def num_qubits(self):
        return 4

    def _eigen_components(self):
        # projector onto subspace spanned by basis states with
        # Hamming weight != 2
        zero_component = np.diag(
            [int(bin(i).count('1') != 2) for i in range(16)])

        state_pairs = (('0110', '1001'), ('0101', '1010'), ('0011', '1100'))

        plus_minus_components = tuple(
            (-abs(weight) * sign / np.pi,
             state_swap_eigen_component(state_pair[0], state_pair[1], sign,
                                        np.angle(weight)))
            for weight, state_pair in zip(self.weights, state_pairs)
            for sign in (-1, 1))

        return ((0, zero_component),) + plus_minus_components

    def _with_exponent(self, exponent: Union[sympy.Symbol, float]
                      ) -> 'QuarticFermionicSimulationGate':
        gate = QuarticFermionicSimulationGate(self.weights)
        gate._exponent = exponent
        return gate

    def _decompose_(self, qubits):
        """The goal is to effect a rotation around an axis in the XY plane in
        each of three orthogonal 2-dimensional subspaces.

        First, the following basis change is performed:
            0000 ↦ 0001        0001 ↦ 1111
            1111 ↦ 0010        1110 ↦ 1100
                               0010 ↦ 0000
            0110 ↦ 0101        1101 ↦ 0011
            1001 ↦ 0110        0100 ↦ 0100
            1010 ↦ 1001        1011 ↦ 0111
            0101 ↦ 1010        1000 ↦ 1000
            1100 ↦ 1101        0111 ↦ 1011
            0011 ↦ 1110

        Note that for each 2-dimensional subspace of interest, the first two
        qubits are the same and the right two qubits are different. The desired
        rotations thus can be effected by a complex-version of a partial SWAP
        gate on the latter two qubits, controlled on the first two qubits. This
        partial SWAP-like gate can  be decomposed such that it is parameterized
        solely by a rotation in the ZY plane on the third qubit. These are the
        `individual_rotations`; call them U0, U1, U2.

        To decompose the double controlled rotations, we use four other
        rotations V0, V1, V2, V3 (the `combined_rotations`) such that
            U0 = V3 · V1 · V0
            U1 = V3 · V2 · V1
            U2 = V2 · V0
        """

        if self._is_parameterized_():
            return NotImplemented

        individual_rotations = [
            la.expm(0.5j * self.exponent *
                    np.array([[np.real(w), 1j * s * np.imag(w)],
                              [-1j * s * np.imag(w), -np.real(w)]]))
            for s, w in zip([1, -1, -1], self.weights)
        ]

        combined_rotations = {}
        combined_rotations[0] = la.sqrtm(
            np.linalg.multi_dot([
                la.inv(individual_rotations[1]), individual_rotations[0],
                individual_rotations[2]
            ]))
        combined_rotations[1] = la.inv(combined_rotations[0])
        combined_rotations[2] = np.linalg.multi_dot([
            la.inv(individual_rotations[0]), individual_rotations[1],
            combined_rotations[0]
        ])
        combined_rotations[3] = individual_rotations[0]

        controlled_rotations = {
            i: cirq.ControlledGate(
                cirq.MatrixGate(combined_rotations[i], qid_shape=(2,)))
            for i in range(4)
        }

        a, b, c, d = qubits

        basis_change = list(
            cirq.flatten_op_tree([
                cirq.CNOT(b, a),
                cirq.CNOT(c, b),
                cirq.CNOT(d, c),
                cirq.CNOT(c, b),
                cirq.CNOT(b, a),
                cirq.CNOT(a, b),
                cirq.CNOT(b, c),
                cirq.CNOT(a, b),
                [cirq.X(c), cirq.X(d)],
                [cirq.CNOT(c, d), cirq.CNOT(d, c)],
                [cirq.X(c), cirq.X(d)],
            ]))

        controlled_rotations = list(
            cirq.flatten_op_tree([
                controlled_rotations[0](b, c),
                cirq.CNOT(a, b), controlled_rotations[1](b, c),
                cirq.CNOT(b, a),
                cirq.CNOT(a, b), controlled_rotations[2](b, c),
                cirq.CNOT(a, b), controlled_rotations[3](b, c)
            ]))

        controlled_swaps = [
            [cirq.CNOT(c, d), cirq.H(c)],
            cirq.CNOT(d, c),
            controlled_rotations,
            cirq.CNOT(d, c),
            [cirq.inverse(op) for op in reversed(controlled_rotations)],
            [cirq.H(c), cirq.CNOT(c, d)],
        ]

        return [basis_change, controlled_swaps, basis_change[::-1]]

    @classmethod
    def wire_symbol(cls, use_unicode: bool):
        return '⇊⇈' if use_unicode else 'a*a*aa'

    def _apply_unitary_(self,
                        args: cirq.ApplyUnitaryArgs) -> Optional[np.ndarray]:
        if cirq.is_parameterized(self):
            return NotImplemented

        am, bm, cm = (la.expm(-1j * self.exponent *
                              np.array([[0, w], [w.conjugate(), 0]]))
                      for w in self.weights)

        a1 = args.subspace_index(0b1001)
        b1 = args.subspace_index(0b0101)
        c1 = args.subspace_index(0b0011)

        a2 = args.subspace_index(0b0110)
        b2 = args.subspace_index(0b1010)
        c2 = args.subspace_index(0b1100)

        cirq.apply_matrix_to_slices(args.target_tensor,
                                    am,
                                    slices=[a1, a2],
                                    out=args.available_buffer)
        cirq.apply_matrix_to_slices(args.available_buffer,
                                    bm,
                                    slices=[b1, b2],
                                    out=args.target_tensor)
        return cirq.apply_matrix_to_slices(args.target_tensor,
                                           cm,
                                           slices=[c1, c2],
                                           out=args.available_buffer)

    def __repr__(self):
        return ('openfermion.QuarticFermionicSimulationGate(({}), '
                'absorb_exponent=False, '
                'exponent={})'.format(
                    ', '.join(
                        cirq._compat.proper_repr(v) for v in self.weights),
                    cirq._compat.proper_repr(self.exponent)))

    @property
    def qubit_generator_matrix(self):
        """The (Hermitian) matrix G such that the gate's unitary is
        exp(-i * G).
        """

        generator = np.zeros((1 << 4,) * 2, dtype=np.complex128)

        # w0 |1001><0110| + h.c.
        generator[9, 6] = self.weights[0]
        generator[6, 9] = self.weights[0].conjugate()
        # w1 |1010><0101| + h.c.
        generator[10, 5] = self.weights[1]
        generator[5, 10] = self.weights[1].conjugate()
        # w2 |1100><0011| + h.c.
        generator[12, 3] = self.weights[2]
        generator[3, 12] = self.weights[2].conjugate()
        return generator

    @staticmethod
    def fermion_generator_components():
        return (
            ops.FermionOperator(((0, 1), (1, 0), (2, 0), (3, 1)), -1),
            ops.FermionOperator(((0, 1), (1, 0), (2, 1), (3, 0))),
            ops.FermionOperator(((0, 1), (1, 1), (2, 0), (3, 0)), -1),
        )

    @classmethod
    def from_interaction_operator(
            cls,
            *,
            operator: 'openfermion.InteractionOperator',
            modes: Optional[Sequence[int]] = None,
    ) -> Optional['QuarticFermionicSimulationGate']:
        if modes is None:
            modes = (0, 1, 2, 3)
        i, j, k, l = modes
        weights = tuple(
            (operator.two_body_tensor[p, q, r, s] -
             operator.two_body_tensor[p, q, s, r] -
             operator.two_body_tensor[q, p, r, s] +
             operator.two_body_tensor[q, p, s, r])
            for p, q, r, s in [(i, l, j, k), (i, k, j, l), (i, j, k, l)])
        if any(weights):
            return cls(cast(Tuple[complex, complex, complex], weights))
        return None

    def interaction_operator_generator(
            self,
            operator: Optional['openfermion.InteractionOperator'] = None,
            modes: Optional[Sequence[int]] = None
    ) -> 'openfermion.InteractionOperator':
        if modes is None:
            modes = (0, 1, 2, 3)
        if operator is None:
            n_modes = max(modes) + 1
            operator = ops.InteractionOperator.zero(n_modes)

        weights = tuple(w * self._exponent for w in self.weights)
        operator.constant += self._exponent * self._global_shift
        p, q, r, s = modes
        operator.two_body_tensor[p, s, q, r] += weights[0]
        operator.two_body_tensor[q, r, p, s] += weights[0].conjugate()
        operator.two_body_tensor[p, r, q, s] += weights[1]
        operator.two_body_tensor[q, s, p, r] += weights[1].conjugate()
        operator.two_body_tensor[p, q, r, s] += weights[2]
        operator.two_body_tensor[r, s, p, q] += weights[2].conjugate()
        return operator

    def fswap(self, i: int):
        if i == 0:
            self.weights = (self.weights[1].conjugate(),
                            self.weights[0].conjugate(), -self.weights[2])
        elif i == 1:
            self.weights = (-self.weights[0], self.weights[2], self.weights[1])
        elif i == 2:
            self.weights = (self.weights[1], self.weights[0], -self.weights[2])
        else:
            raise ValueError(f'{i} not in (0, 1, 2)')



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import io
import os

from setuptools import setup, find_packages


# This reads the __version__ variable from openfermion/_version.py
__version__ = ''
exec(open('src/openfermion/_version.py').read())

# Readme file as long_description:
long_description = ('===========\n' +
                    'OpenFermion\n' +
                    '===========\n')
stream = io.open('README.rst', encoding='utf-8')
stream.readline()
long_description += stream.read()

# Read in package requirements.txt
requirements = open('requirements.txt').readlines()
requirements = [r.strip() for r in requirements]

docs_files_gen = os.walk('docs')
docs_data_files_tuples = []
for cwd, subdirs, files in list(docs_files_gen)[1:]:
    if 'ipynb_checkpoints' in cwd:
        continue
    docs_data_files_tuples.append(
        (os.path.join('openfermion',
                      cwd), [os.path.join(cwd, file) for file in files]))

setup(
    name='openfermion',
    version=__version__,
    author='The OpenFermion Developers',
    author_email='help@openfermion.org',
    url='http://www.openfermion.org',
    description=('The electronic structure package for quantum computers.'),
    long_description=long_description,
    install_requires=requirements,
    license='Apache 2',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    package_data={
        '': [
            os.path.join('src', 'openfermion', 'testing', '*.npy'),
            os.path.join('src', 'openfermion', 'testing', '*.hdf5'),
        ],
    },
    data_files=docs_data_files_tuples,
)



import numpy as np
from openfermion.linalg.rdm_reconstruction import valdemoro_reconstruction
from openfermion.linalg.wedge_product import wedge


def test_valdemoro_correct_size():
    m = 4
    opdm = np.random.random((m, m))
    tpdm = wedge(opdm, opdm, (1, 1), (1, 1))
    d3v = valdemoro_reconstruction(tpdm, 4)
    assert d3v.shape == (m, m, m, m, m, m)


def test_correct_trace():
    opdm = np.diag([1] * 2 + [0] * 2)
    tpdm = wedge(opdm, opdm, (1, 1), (1, 1))
    assert np.isclose(np.einsum('ijji', tpdm), 2 * (2 - 1) / 2)
    d3v = valdemoro_reconstruction(tpdm, 2)
    # because only 2 particles in this system
    assert np.isclose(np.einsum('ijkkji', d3v), 0)

    opdm = np.diag([1] * 5 + [0] * 3)
    tpdm = wedge(opdm, opdm, (1, 1), (1, 1))
    assert np.isclose(np.einsum('ijji', tpdm), 5 * (5 - 1) / 2)
    d3v = valdemoro_reconstruction(tpdm, 5)
    assert np.isclose(np.einsum('ijkkji', d3v), 5 * (5 - 1) * (5 - 2) / 6)